# -*- coding: utf-8 -*-
#
# Copyright 2019 SoloKeys Developers
#
# Licensed under the Apache License, Version 2.0, <LICENSE-APACHE or
# http://apache.org/licenses/LICENSE-2.0> or the MIT license <LICENSE-MIT or
# http://opensource.org/licenses/MIT>, at your option. This file may not be
# copied, modified, or distributed except according to those terms.

import base64
import binascii
import hashlib
import json
import secrets
import struct
import sys
import tempfile
import time
from getpass import getpass
from typing import Any, Callable, List, Optional, Tuple, Union

from fido2.client import Fido2Client, UserInteraction
from fido2.cose import ES256, EdDSA
from fido2.ctap import CtapError
from fido2.ctap1 import Ctap1
from fido2.ctap2.base import Ctap2
from fido2.ctap2.credman import CredentialManagement
from fido2.ctap2.pin import ClientPin
from fido2.hid import CTAPHID, CtapHidDevice, open_device
from fido2.webauthn import (
    AuthenticatorSelectionCriteria,
    PublicKeyCredentialCreationOptions,
    PublicKeyCredentialParameters,
    PublicKeyCredentialRpEntity,
    PublicKeyCredentialType,
    PublicKeyCredentialUserEntity,
    ResidentKeyRequirement,
    UserVerificationRequirement,
)
from intelhex import IntelHex

import pynitrokey.exceptions
import pynitrokey.fido2 as nkfido2
from pynitrokey import helpers
from pynitrokey.fido2.commands import SoloBootloader, SoloExtension
from pynitrokey.helpers import local_critical, local_print


class CliOrProvidedInteraction(UserInteraction):
    def __init__(self, pin: Optional[str]) -> None:
        self.pin = pin

    def prompt_up(self) -> None:
        print("Touch your authenticator device now...")

    def request_pin(self, permissions: Any, rd_id: Any) -> str:
        if self.pin:
            return self.pin
        else:
            return getpass("Enter PIN: ")

    def request_uv(self, permissions: Any, rd_id: Any) -> bool:
        return True


class NKFido2Client:
    def __init__(self) -> None:
        self.origin = "https://example.org"
        self.host = "example.org"
        self.user_id = b"they"
        self.exchange = self.exchange_hid
        self.do_reboot = True

    def use_u2f(self) -> None:
        self.exchange = self.exchange_u2f

    def use_hid(self) -> None:
        self.exchange = self.exchange_hid

    def set_reboot(self, val: bool) -> None:
        """option to reboot after programming"""
        self.do_reboot = val

    def reboot(self) -> None:
        """option to reboot after programming"""
        try:
            self.exchange(SoloBootloader.reboot)
        except OSError:
            pass

    def find_device(
        self,
        dev: Optional[CtapHidDevice] = None,
        solo_serial: Optional[str] = None,
        pin: Optional[str] = None,
    ) -> CtapHidDevice:

        devices = []
        found_dev: Optional[CtapHidDevice] = None

        if dev is None:
            if solo_serial is not None:
                if solo_serial.startswith("device="):
                    solo_serial = solo_serial.split("=")[1]
                    found_dev = open_device(solo_serial)
                else:
                    devices = list(CtapHidDevice.list_devices())
                    devices = [
                        d for d in devices if d.descriptor.serial_number == solo_serial
                    ]
            else:
                devices = list(CtapHidDevice.list_devices())
            if len(devices) > 1:
                raise pynitrokey.exceptions.NonUniqueDeviceError
            if len(devices) > 0:
                found_dev = devices[0]
        if dev is None and found_dev is None:
            raise RuntimeError("No FIDO device found")

        self.dev = found_dev if found_dev else dev
        assert isinstance(self.dev, CtapHidDevice)
        self.ctap1 = Ctap1(self.dev)

        try:
            self.ctap2: Optional[Ctap2] = Ctap2(self.dev)
        except CtapError as e:
            self.ctap2 = None

        try:
            self.client: Optional[Fido2Client] = Fido2Client(
                self.dev, self.origin, user_interaction=CliOrProvidedInteraction(pin)
            )
        except CtapError:
            print("Not using FIDO2 interface.")
            self.client = None

        if self.exchange == self.exchange_hid:
            self.send_data_hid(CTAPHID.INIT, b"\x11\x11\x11\x11\x11\x11\x11\x11")

        return self.dev

    @staticmethod
    def format_request(cmd: int, addr: int = 0, data: bytes = b"A" * 16) -> bytes:
        # not sure why this is here?
        # arr = b"\x00" * 9
        packed_addr = struct.pack("<L", addr)
        packed_cmd = struct.pack("B", cmd)
        length = struct.pack(">H", len(data))

        return packed_cmd + packed_addr[:3] + SoloBootloader.TAG + length + data

    def send_only_hid(self, cmd: int, data: bytes = b"A" * 16) -> None:
        if not isinstance(data, bytes):
            data = struct.pack("%dB" % len(data), *[ord(x) for x in data])
        assert isinstance(self.dev, CtapHidDevice)
        self.dev.call(0x80 | cmd, bytearray(data))

    def send_data_hid(self, cmd: int, data: bytes = b"A" * 16) -> bytes:
        if not isinstance(data, bytes):
            data = struct.pack("%dB" % len(data), *[ord(x) for x in data])
        with helpers.Timeout(1.0) as event:
            assert isinstance(self.dev, CtapHidDevice)
            return self.dev.call(cmd, data, event=event)

    def exchange_hid(self, cmd: int, addr: int = 0, data: bytes = b"A" * 16) -> bytes:
        req = NKFido2Client.format_request(cmd, addr, data)

        data = self.send_data_hid(SoloBootloader.HIDCommandBoot, req)

        ret = data[0]
        if ret != CtapError.ERR.SUCCESS:
            raise CtapError(ret)

        return data[1:]

    def exchange_u2f(self, cmd: int, addr: int = 0, data: bytes = b"A" * 16) -> bytes:
        appid = b"A" * 32
        chal = b"B" * 32

        req = NKFido2Client.format_request(cmd, addr, data)

        res = self.ctap1.authenticate(chal, appid, req)

        ret = res.signature[0]
        if ret != CtapError.ERR.SUCCESS:
            raise CtapError(ret)

        return res.signature[1:]

    def exchange_fido2(self, cmd: int, addr: int = 0, data: bytes = b"A" * 16) -> bytes:
        chal = b"B" * 32

        req = NKFido2Client.format_request(cmd, addr, data)

        assert isinstance(self.ctap2, Ctap2)
        assertion = self.ctap2.get_assertion(
            self.host, chal, [{"id": req, "type": "public-key"}]
        )

        res = assertion
        ret = res.signature[0]
        if ret != CtapError.ERR.SUCCESS:
            raise RuntimeError("Device returned non-success code %02x" % (ret,))

        return res.signature[1:]

    def bootloader_version(self) -> Tuple[int, int, int]:
        data = self.exchange(SoloBootloader.version)
        if len(data) > 2:
            return (data[0], data[1], data[2])
        return (0, 0, data[0])

    def solo_version(self) -> Union[bytes, Tuple[int, int, int]]:
        try:
            return self.send_data_hid(0x61, b"")
        except CtapError:
            data = self.exchange(SoloExtension.version)
            return (data[0], data[1], data[2])

    def write_flash(self, addr: int, data: bytes) -> None:
        self.exchange(SoloBootloader.write, addr, data)

    def boot_pubkey(self) -> bytes:
        return self.exchange(SoloBootloader.boot_pubkey)

    def get_rng(self, num: int = 0) -> bytes:
        ret = self.send_data_hid(SoloBootloader.HIDCommandRNG, struct.pack("B", num))
        return ret

    def get_status(self, num: int = 0) -> bytes:
        ret = self.send_data_hid(SoloBootloader.HIDCommandStatus, struct.pack("B", num))
        # print(ret[:8])
        return ret

    def verify_flash(self, sig: bytes) -> None:
        """
        Tells device to check signature against application.  If it passes,
        the application will boot.
        Exception raises if signature fails.
        """
        self.exchange(SoloBootloader.done, 0, sig)

    def wink(self) -> None:
        self.send_data_hid(CTAPHID.WINK, b"")

    def reset(self) -> None:
        assert isinstance(self.ctap2, Ctap2)
        self.ctap2.reset()

    def make_credential(
        self,
        host: str = "nitrokeys.dev",
        user_id: str = "they",
        resident_key: str = "",
        user_verification: str = "",
        output: bool = True,
        fingerprint_only: bool = False,
    ) -> str:

        """
        fingerprint_only bool Return sha256 digest of the certificate, in a hex string format. Useful for detecting
            device's model and firmware.
        """

        assert self.client is not None

        options = PublicKeyCredentialCreationOptions(
            rp=PublicKeyCredentialRpEntity(name="Example RP", id=host),
            user=PublicKeyCredentialUserEntity(name="A. User", id=user_id.encode()),
            challenge=secrets.token_bytes(32),
            pub_key_cred_params=[
                PublicKeyCredentialParameters(
                    type=PublicKeyCredentialType.PUBLIC_KEY, alg=EdDSA.ALGORITHM
                ),
                PublicKeyCredentialParameters(
                    type=PublicKeyCredentialType.PUBLIC_KEY, alg=ES256.ALGORITHM
                ),
            ],
            extensions={"hmacCreateSecret": True},
            authenticator_selection=AuthenticatorSelectionCriteria(
                resident_key=ResidentKeyRequirement(resident_key),
                user_verification=UserVerificationRequirement(user_verification),
            ),
        )
        self.client.origin = f"https://{host}"
        attestation_object = self.client.make_credential(options).attestation_object

        if fingerprint_only:
            if "x5c" not in attestation_object.att_stmt:
                raise ValueError("No x5c information available")
            from hashlib import sha256

            data = attestation_object.att_stmt["x5c"]
            return sha256(data[0]).digest().hex()

        credential = attestation_object.auth_data.credential_data
        if not credential:
            raise ValueError("No credential ID available")
        credential_id = credential.credential_id
        if output:
            print(credential_id.hex())

        return credential_id.hex()

    def simple_secret(
        self,
        credential_id: str,
        secret_input: str,
        host: str = "nitrokeys.dev",
        user_id: str = "they",
        serial: Optional[str] = None,
        prompt: Optional[str] = "Touch your authenticator to generate a response...",
        output: bool = True,
        udp: bool = False,
    ) -> bytes:

        _user_id = user_id.encode()

        assert isinstance(self.client, Fido2Client)
        client = self.client

        # @todo: rewrite with typing; use fido2.webauthn.PublicKeyCredentialCreationOptions
        # rp = {"id": host, "name": "Example RP"}
        client.host = host  # type: ignore
        client.origin = f"https://{client.host}"  # type: ignore
        client.user_id = _user_id  # type: ignore
        # user = {"id": user_id, "name": "A. User"}
        _credential_id = binascii.a2b_hex(credential_id)

        allow_list = [{"type": "public-key", "id": _credential_id}]

        challenge = secrets.token_bytes(32)

        h = hashlib.sha256()
        h.update(secret_input.encode())
        salt = h.digest()

        if prompt:
            print(prompt)

        assertion = client.get_assertion(
            {
                "rpId": host,
                "challenge": challenge,
                "allowCredentials": allow_list,
                "extensions": {"hmacGetSecret": {"salt1": salt}},
            },  # type: ignore
        ).get_response(0)

        # @todo: rewrite with typing
        output = assertion.extension_results["hmacGetSecret"]["output1"]  # type: ignore
        assert isinstance(output, bytes)
        if output:
            print(output.hex())

        return output

    def has_pin(self) -> bool:
        assert self.client is not None
        return self.client.info.options["clientPin"]

    def cred_mgmt(self, serial: str, pin: str) -> CredentialManagement:
        device = nkfido2.find(serial)
        assert isinstance(device.ctap2, Ctap2)
        client_pin = ClientPin(device.ctap2)

        try:
            client_token = client_pin.get_pin_token(pin)
        except CtapError as error:
            if error.code == CtapError.ERR.PIN_NOT_SET:
                local_critical("Please set a pin in order to manage credentials")
            if error.code == CtapError.ERR.PIN_AUTH_BLOCKED:
                local_critical(
                    "Pin authentication has been blocked, try reinserting the key or setting a pin if none is set"
                )
            if error.code == CtapError.ERR.PIN_BLOCKED:
                local_critical(
                    "Your device has been blocked after too many failed unlock attempts, to fix this it "
                    "will have to be reset. (If no pin is set, plugging it in again might fix this warning)"
                )
            raise

        return CredentialManagement(device.ctap2, client_pin.protocol, client_token)

    def enter_bootloader(self) -> None:
        """
        If Nitrokey is configured as Nitrokey hacker or something similar,
        this command will tell the token to boot directly to the bootloader
        so it can be reprogrammed
        """
        if self.exchange != self.exchange_hid:
            self.send_data_hid(CTAPHID.INIT, b"\x11\x11\x11\x11\x11\x11\x11\x11")
        self.send_data_hid(SoloBootloader.HIDCommandEnterBoot, b"")

    def enter_bootloader_or_die(self) -> None:
        try:
            self.enter_bootloader()
        # except OSError:
        #     pass
        except CtapError as e:
            if e.code == CtapError.ERR.INVALID_COMMAND:
                print(
                    "Could not switch into bootloader mode.  Please hold down the button for 2s while you plug token in."
                )
                sys.exit(1)
            else:
                raise (e)

    def is_bootloader(self) -> bool:
        try:
            self.bootloader_version()
            return True
        except CtapError as e:
            if e.code == CtapError.ERR.INVALID_COMMAND:
                pass
            else:
                raise (e)
        except Exception as e:
            # exception during bootloader version check, assume no bootloader
            # local_print("could not get bootloader version")
            pass
        return False

    def enter_st_dfu(self) -> None:
        """
        If Nitrokey is configured as Nitrokey hacker or something similar,
        this command will tell the token to boot directly to the st DFU
        so it can be reprogrammed.  Warning, you could brick your device.
        """
        boot = self.is_bootloader()

        if boot or self.exchange == self.exchange_u2f:
            req = NKFido2Client.format_request(SoloBootloader.st_dfu)
            self.send_only_hid(SoloBootloader.HIDCommandBoot, req)
        else:
            self.send_only_hid(SoloBootloader.HIDCommandEnterSTBoot, b"")

    def program_file(self, name: str) -> bytes:
        def parseField(f: str) -> bytes:
            return base64.b64decode(helpers.from_websafe(f).encode())

        def isCorrectVersion(current: Tuple[int, int, int], target: str) -> bool:
            """current is tuple (x,y,z).  target is string '>=x.y.z'.
            Return True if current satisfies the target expression.
            """
            if "=" in target:
                target_toks = target.split("=")
                assert target_toks[0] in [">", "<"]
                target_num_toks = [int(x) for x in target_toks[1].split(".")]
                assert len(target_num_toks) == 3
                comp = target_toks[0] + "="
            else:
                assert target[0] in [">", "<"]
                target_num_toks = [int(x) for x in target[1:].split(".")]
                comp = target[0]
            target_num = (
                (target_num_toks[0] << 16)
                | (target_num_toks[1] << 8)
                | (target_num_toks[2] << 0)
            )
            current_num = (current[0] << 16) | (current[1] << 8) | (current[2] << 0)
            return eval(str(current_num) + comp + str(target_num))  # type: ignore [no-any-return]

        firmware_file_data = None
        if name.lower().endswith(".json"):
            firmware_file_data = json.loads(open(name, "r").read())
            fw = parseField(firmware_file_data["firmware"])
            sig = None

            if "versions" in firmware_file_data:
                current = (0, 0, 0)
                try:
                    current = self.bootloader_version()
                except CtapError as e:
                    if e.code == CtapError.ERR.INVALID_COMMAND:
                        pass
                    else:
                        raise (e)
                # for v in firmware_file_data["versions"]:
                #     if not isCorrectVersion(current, v):
                #         print("using signature version", v)
                #         sig = parseField(firmware_file_data["versions"][v]["signature"])
                #         break
                sig = parseField(firmware_file_data["versions"][">2.5.3"]["signature"])

                if sig is None:
                    raise RuntimeError(
                        "Improperly formatted firmware file.  Could not match version."
                    )
            else:
                sig = parseField(firmware_file_data["signature"])

            ih = IntelHex()
            tmp = tempfile.NamedTemporaryFile(delete=False)
            tmp.write(fw)
            tmp.seek(0)
            tmp.close()
            ih.fromfile(tmp.name, format="hex")
        else:
            if not name.lower().endswith(".hex"):
                print('Warning, assuming "%s" is an Intel Hex file.' % name)
            sig = None
            ih = IntelHex()
            ih.fromfile(name, format="hex")

        if self.exchange == self.exchange_hid:
            chunk = 2048
        else:
            chunk = 240

        seg = ih.segments()[0]
        size = seg[1] - seg[0]
        total = 0
        t1 = time.time() * 1000
        print("erasing firmware...")
        for i in range(seg[0], seg[1], chunk):
            s = i
            ext = min(i + chunk, seg[1])
            data = ih.tobinarray(start=i, size=ext - s)
            self.write_flash(i, data)
            total += chunk
            progress = total / float(size) * 100
            sys.stdout.write("updating firmware %.2f%%...\r" % progress)
        sys.stdout.write("updated firmware 100%             \r\n")
        t2 = time.time() * 1000
        print("time: %.2f s" % ((t2 - t1) / 1000.0))

        if sig is None:
            sig = b"A" * 64

        success = False
        if self.do_reboot:
            try:
                print("bootloader is verifying signature...")
                print(f"Trying with {sig.hex()}")
                self.verify_flash(sig)
                print("...pass!")
                success = True
            except CtapError as e:
                if e.code != 0x27:
                    raise
                print("...error!")

        if not success:
            msg = """Bootloader reports failure in the signature verification. If your device is staying in the
            bootloader mode after reinserting it into the USB port, please execute the following to make it work again:

            # Download the compressed firmware file for Nitrokey FIDO2 128 v2.4.1 and extract it
            wget https://github.com/Nitrokey/nitrokey-fido2-firmware/releases/download/2.4.1.nitrokey/nitrokey-fido2-firmware-2.4.1-128kB-app-signed.zip
            unzip nitrokey-fido2-firmware-2.4.1-128kB-app-signed.zip
            # Run the update process again with the just downloaded firmware
            nitropy fido2 util program bootloader nitrokey-fido2-firmware-2.4.1-128kB-app-signed.json
            """
            local_critical(msg, support_hint=False)

        return sig

    def check_only(self, name: str) -> None:
        # FIXME refactor
        # copy from program_file
        if name.lower().endswith(".json"):
            data = json.loads(open(name, "r").read())
            fw = base64.b64decode(helpers.from_websafe(data["firmware"]).encode())
            sig = base64.b64decode(helpers.from_websafe(data["signature"]).encode())
            ih = IntelHex()
            tmp = tempfile.NamedTemporaryFile(delete=False)
            tmp.write(fw)
            tmp.seek(0)
            tmp.close()
            ih.fromfile(tmp.name, format="hex")
        else:
            if not name.lower().endswith(".hex"):
                print('Warning, assuming "%s" is an Intel Hex file.' % name)
            sig = None
            ih = IntelHex()
            ih.fromfile(name, format="hex")

        if sig is None:
            sig = b"A" * 64

        if self.do_reboot:
            self.verify_flash(sig)



"""
gnuk_token.py - a library for Gnuk Token

Copyright (C) 2011, 2012, 2013, 2015, 2017, 2018
              Free Software Initiative of Japan
Author: NIIBE Yutaka <gniibe@fsij.org>

This file is a part of Gnuk, a GnuPG USB Token implementation.

Gnuk is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Gnuk is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import binascii
import logging
import time
from array import array
from contextlib import contextmanager
from struct import *
from typing import Optional

import usb

# Possible Gnuk Token products
from pynitrokey.start.usb_strings import get_dict_for_device

USB_PRODUCT_LIST = [
    {"vendor": 0x234B, "product": 0x0000},  # FSIJ Gnuk Token
    {"vendor": 0x20A0, "product": 0x4211},  # Nitrokey Start
    {"vendor": 0x1209, "product": 0x2440},  # GnuPG e.V.
]

USB_PRODUCT_LIST_TUP = [
    (0x234B, 0x0000),  # FSIJ Gnuk Token
    (0x20A0, 0x4211),  # Nitrokey Start
    (0x1209, 0x2440),  # GnuPG e.V.
]

# USB class, subclass, protocol
CCID_CLASS = 0x0B
CCID_SUBCLASS = 0x00
CCID_PROTOCOL_0 = 0x00

HID_CLASS = 0x03
HID_SUBCLASS_NO_BOOT = 0x00
HID_PROTOCOL_0 = 0x00


def icc_compose(msg_type, data_len, slot, seq, param, data):
    return pack("<BiBBBH", msg_type, data_len, slot, seq, 0, param) + data


def iso7816_compose(
    ins: int,
    p1: int,
    p2: int,
    data: bytes = b"",
    cls: int = 0x00,
    le: Optional[int] = None,
) -> bytes:
    data_len = len(data)
    if data_len == 0:
        if not le:
            return pack(">BBBB", cls, ins, p1, p2)
        else:
            return pack(">BBBBB", cls, ins, p1, p2, le)
    else:
        if data_len <= 255:
            if not le:
                return pack(">BBBBB", cls, ins, p1, p2, data_len) + data
            else:
                return (
                    pack(">BBBBB", cls, ins, p1, p2, data_len) + data + pack(">B", le)
                )
        else:
            if not le:
                return pack(">BBBBBH", cls, ins, p1, p2, 0, data_len) + data
            else:
                return (
                    pack(">BBBBBH", cls, ins, p1, p2, 0, data_len)
                    + data
                    + pack(">B", le)
                )


# This class only supports Gnuk (for now)
class gnuk_token(object):
    def __init__(self, device, configuration, interface):
        """
        __init__(device, configuration, interface) -> None
        Initialize the device.
        device: usb.Device object.
        configuration: configuration number.
        interface: usb.Interface object representing the interface and altenate setting.
        """
        if interface.interfaceClass != CCID_CLASS:
            raise ValueError("Wrong interface class")
        if interface.interfaceSubClass != CCID_SUBCLASS:
            raise ValueError("Wrong interface sub class")
        self.__devhandle = device.open()
        self.__devhandle.claimInterface(interface)
        self.__devhandle.setAltInterface(interface)

        self.__intf = interface.interfaceNumber
        self.__alt = interface.alternateSetting
        self.__conf = configuration

        self.__hid_intf = None
        for intf in configuration.interfaces:
            for alt in intf:
                if (
                    alt.interfaceClass == HID_CLASS
                    and alt.interfaceSubClass == HID_SUBCLASS_NO_BOOT
                    and alt.interfaceProtocol == HID_PROTOCOL_0
                ):
                    self.__hid_intf = alt.interfaceNumber

        self.__bulkout = 1
        self.__bulkin = 0x81

        self.__timeout = 10000
        self.__seq = 0
        self.logger = logging.getLogger("gnuk_token")

    def set_logger(self, logger: logging.Logger):
        self.logger = logger.getChild("gnuk_token")

    def local_print(self, message: str, verbose=False):
        self.logger.debug("print: {}".format(message))
        if verbose:
            print(message)

    def get_string(self, num):
        return self.__devhandle.getString(num, 512)

    def increment_seq(self):
        self.__seq = (self.__seq + 1) & 0xFF

    def reset_device(self):
        try:
            self.__devhandle.reset()
        except:
            pass

    def release_gnuk(self):
        self.__devhandle.releaseInterface()

    @contextmanager
    def release_on_exit(self, *args, **kwds):
        """
        Release GNUK device once finished, to not block other smart card services
        trying to claim it
        """
        try:
            yield self
        finally:
            self.release_gnuk()

    def stop_gnuk(self):
        self.__devhandle.releaseInterface()
        if self.__hid_intf:
            self.__devhandle.detachKernelDriver(self.__hid_intf)
        self.__devhandle.setConfiguration(0)
        return

    def mem_info(self):
        mem = self.__devhandle.controlMsg(
            requestType=0xC0, request=0, buffer=8, value=0, index=0, timeout=10
        )
        start = ((mem[3] * 256 + mem[2]) * 256 + mem[1]) * 256 + mem[0]
        end = ((mem[7] * 256 + mem[6]) * 256 + mem[5]) * 256 + mem[4]
        return (start, end)

    def download(self, start, data, verbose=False, progress_func=None):
        addr = start
        addr_end = (start + len(data)) & 0xFFFFFF00
        i = int((addr - 0x20000000) / 0x100)
        j = 0
        self.local_print("start %08x" % addr, verbose)
        self.local_print("end   %08x" % addr_end)
        if progress_func:
            progress_func(0)
        while addr < addr_end:
            if progress_func:
                progress_func((addr - start) / (addr_end - start))
            self.local_print("# %08x: %d : %d" % (addr, i, 256), verbose)
            self.__devhandle.controlMsg(
                requestType=0x40,
                request=1,
                buffer=data[j * 256 : j * 256 + 256],
                value=i,
                index=0,
                timeout=10,
            )
            i = i + 1
            j = j + 1
            addr = addr + 256
        residue = len(data) % 256
        if residue != 0:
            self.local_print("# %08x: %d : %d" % (addr, i, residue), verbose)
            self.__devhandle.controlMsg(
                requestType=0x40,
                request=1,
                buffer=data[j * 256 :],
                value=i,
                index=0,
                timeout=10,
            )

    def execute(self, last_addr):
        i = int((last_addr - 0x20000000) / 0x100)
        o = (last_addr - 0x20000000) % 0x100
        self.__devhandle.controlMsg(
            requestType=0x40, request=2, buffer=None, value=i, index=o, timeout=10
        )

    def icc_get_result(self):
        usbmsg = self.__devhandle.bulkRead(self.__bulkin, 1024, self.__timeout)
        if len(usbmsg) < 10:
            self.local_print(usbmsg, True)
            raise ValueError("icc_get_result")
        msg = array("B", usbmsg)
        msg_type = msg[0]
        data_len = msg[1] + (msg[2] << 8) + (msg[3] << 16) + (msg[4] << 24)
        slot = msg[5]
        seq = msg[6]
        status = msg[7]
        error = msg[8]
        chain = msg[9]
        data = msg[10:]
        # XXX: check msg_type, data_len, slot, seq, error
        return (status, chain, data)

    def icc_get_status(self):
        msg = icc_compose(0x65, 0, 0, self.__seq, 0, b"")
        self.__devhandle.bulkWrite(self.__bulkout, msg, self.__timeout)
        self.increment_seq()
        status, chain, data = self.icc_get_result()
        # XXX: check chain, data
        return status

    def icc_power_on(self):
        msg = icc_compose(0x62, 0, 0, self.__seq, 0, b"")
        self.__devhandle.bulkWrite(self.__bulkout, msg, self.__timeout)
        self.increment_seq()
        status, chain, data = self.icc_get_result()
        # XXX: check status, chain
        self.atr = data
        return self.atr

    def icc_power_off(self):
        msg = icc_compose(0x63, 0, 0, self.__seq, 0, b"")
        self.__devhandle.bulkWrite(self.__bulkout, msg, self.__timeout)
        self.increment_seq()
        status, chain, data = self.icc_get_result()
        # XXX: check chain, data
        return status

    def icc_send_data_block(self, data):
        msg = icc_compose(0x6F, len(data), 0, self.__seq, 0, data)
        self.__devhandle.bulkWrite(self.__bulkout, msg, self.__timeout)
        self.increment_seq()
        return self.icc_get_result()

    def icc_send_cmd(self, data):
        status, chain, data_rcv = self.icc_send_data_block(data)
        if chain == 0:
            while status == 0x80:
                status, chain, data_rcv = self.icc_get_result()
            return data_rcv
        elif chain == 1:
            d = data_rcv
            while True:
                msg = icc_compose(0x6F, 0, 0, self.__seq, 0x10, b"")
                self.__devhandle.bulkWrite(self.__bulkout, msg, self.__timeout)
                self.increment_seq()
                status, chain, data_rcv = self.icc_get_result()
                # XXX: check status
                d += data_rcv
                if chain == 2:
                    break
                elif chain == 3:
                    continue
                else:
                    raise ValueError("icc_send_cmd chain")
            return d
        else:
            raise ValueError("icc_send_cmd")

    def cmd_get_response(self, expected_len):
        result = array("B")
        while True:
            cmd_data = iso7816_compose(0xC0, 0x00, 0x00, b"") + pack(">B", expected_len)
            response = self.icc_send_cmd(cmd_data)
            result += response[:-2]
            sw = response[-2:]
            if sw[0] == 0x90 and sw[1] == 0x00:
                return result
            elif sw[0] != 0x61:
                raise ValueError("%02x%02x" % (sw[0], sw[1]))
            else:
                expected_len = sw[1]

    def cmd_verify(self, who, passwd):
        cmd_data = iso7816_compose(0x20, 0x00, 0x80 + who, passwd)
        sw = self.icc_send_cmd(cmd_data)
        if len(sw) != 2:
            raise ValueError(sw)
        if not (sw[0] == 0x90 and sw[1] == 0x00):
            raise ValueError("%02x%02x" % (sw[0], sw[1]))
        return True

    def cmd_read_binary(self, fileid):
        cmd_data = iso7816_compose(0xB0, 0x80 + fileid, 0x00, b"")
        sw = self.icc_send_cmd(cmd_data)
        if len(sw) != 2:
            raise ValueError(sw)
        if sw[0] != 0x61:
            raise ValueError("%02x%02x" % (sw[0], sw[1]))
        return self.cmd_get_response(sw[1])

    def cmd_write_binary(self, fileid, data, is_update):
        count = 0
        data_len = len(data)
        if is_update:  # overwrite existing file -> update
            ins = 0xD6
        else:
            ins = 0xD0  # write file, and break if exist already
        while count * 256 < data_len:
            if count == 0:
                if len(data) < 128:
                    cmd_data0 = iso7816_compose(ins, 0x80 + fileid, 0x00, data[:128])
                    cmd_data1 = None
                else:
                    cmd_data0 = iso7816_compose(
                        ins, 0x80 + fileid, 0x00, data[:128], 0x10
                    )
                    cmd_data1 = iso7816_compose(ins, 0x80 + fileid, 0x00, data[128:256])
            else:
                if len(data[256 * count : 256 * count + 128]) < 128:
                    cmd_data0 = iso7816_compose(
                        ins, count, 0x00, data[256 * count : 256 * count + 128]
                    )
                    cmd_data1 = None
                else:
                    cmd_data0 = iso7816_compose(
                        ins, count, 0x00, data[256 * count : 256 * count + 128], 0x10
                    )
                    cmd_data1 = iso7816_compose(
                        ins, count, 0x00, data[256 * count + 128 : 256 * (count + 1)]
                    )
            sw = self.icc_send_cmd(cmd_data0)
            if len(sw) != 2:
                raise ValueError("cmd_write_binary 0")
            if not (sw[0] == 0x90 and sw[1] == 0x00):
                raise ValueError("cmd_write_binary 0", "%02x%02x" % (sw[0], sw[1]))
            if cmd_data1:
                sw = self.icc_send_cmd(cmd_data1)
                if len(sw) != 2:
                    raise ValueError("cmd_write_binary 1", sw)
                if not (sw[0] == 0x90 and sw[1] == 0x00):
                    raise ValueError("cmd_write_binary 1", "%02x%02x" % (sw[0], sw[1]))
            count += 1

    def cmd_select_openpgp(self):
        cmd_data = iso7816_compose(0xA4, 0x04, 0x00, b"\xD2\x76\x00\x01\x24\x01")
        r = self.icc_send_cmd(cmd_data)
        if len(r) < 2:
            raise ValueError(r)
        sw = r[-2:]
        r = r[0:-2]
        if sw[0] == 0x61:
            self.cmd_get_response(sw[1])
            return True
        elif sw[0] == 0x90 and sw[1] == 0x00:
            return True
        else:
            raise ValueError("%02x%02x" % (sw[0], sw[1]))

    def cmd_get_data(self, tagh, tagl):
        cmd_data = iso7816_compose(0xCA, tagh, tagl, b"")
        sw = self.icc_send_cmd(cmd_data)
        if len(sw) != 2:
            raise ValueError(sw)
        if sw[0] == 0x90 and sw[1] == 0x00:
            return array("B")
        elif sw[0] != 0x61:
            raise ValueError("%02x%02x" % (sw[0], sw[1]))
        return self.cmd_get_response(sw[1])

    def cmd_set_identity(self, ident):
        cmd_data = iso7816_compose(0x85, 0x00, ident, b"")
        sw = self.icc_send_cmd(cmd_data)
        if len(sw) != 2:
            raise ValueError(sw)
        if not (sw[0] == 0x90 and sw[1] == 0x00):
            raise ValueError("%02x%02x" % (sw[0], sw[1]))
        return True

    def cmd_change_reference_data(self, who, data):
        cmd_data = iso7816_compose(0x24, 0x00, 0x80 + who, data)
        sw = self.icc_send_cmd(cmd_data)
        if len(sw) != 2:
            raise ValueError(sw)
        if not (sw[0] == 0x90 and sw[1] == 0x00):
            raise ValueError("%02x%02x" % (sw[0], sw[1]))
        return True

    def cmd_put_data(self, tagh, tagl, content):
        cmd_data = iso7816_compose(0xDA, tagh, tagl, content)
        sw = self.icc_send_cmd(cmd_data)
        if len(sw) != 2:
            raise ValueError(sw)
        if not (sw[0] == 0x90 and sw[1] == 0x00):
            raise ValueError("%02x%02x" % (sw[0], sw[1]))
        return True

    def cmd_put_data_odd(self, tagh, tagl, content):
        cmd_data0 = iso7816_compose(0xDB, tagh, tagl, content[:128], 0x10)
        cmd_data1 = iso7816_compose(0xDB, tagh, tagl, content[128:])
        sw = self.icc_send_cmd(cmd_data0)
        if len(sw) != 2:
            raise ValueError(sw)
        if not (sw[0] == 0x90 and sw[1] == 0x00):
            raise ValueError("%02x%02x" % (sw[0], sw[1]))
        sw = self.icc_send_cmd(cmd_data1)
        if len(sw) != 2:
            raise ValueError(sw)
        if not (sw[0] == 0x90 and sw[1] == 0x00):
            raise ValueError("%02x%02x" % (sw[0], sw[1]))
        return True

    def cmd_reset_retry_counter(self, how, who, data):
        cmd_data = iso7816_compose(0x2C, how, who, data)
        sw = self.icc_send_cmd(cmd_data)
        if len(sw) != 2:
            raise ValueError(sw)
        if not (sw[0] == 0x90 and sw[1] == 0x00):
            raise ValueError("%02x%02x" % (sw[0], sw[1]))
        return True

    def cmd_pso(self, p1, p2, data):
        cmd_data = iso7816_compose(0x2A, p1, p2, data)
        sw = self.icc_send_cmd(cmd_data)
        if len(sw) != 2:
            raise ValueError(sw)
        if sw[0] == 0x90 and sw[1] == 0x00:
            return array("B")
        elif sw[0] != 0x61:
            raise ValueError("%02x%02x" % (sw[0], sw[1]))
        return self.cmd_get_response(sw[1])

    def cmd_pso_longdata(self, p1, p2, data):
        cmd_data0 = iso7816_compose(0x2A, p1, p2, data[:128], 0x10)
        cmd_data1 = iso7816_compose(0x2A, p1, p2, data[128:])
        sw = self.icc_send_cmd(cmd_data0)
        if len(sw) != 2:
            raise ValueError(sw)
        if not (sw[0] == 0x90 and sw[1] == 0x00):
            raise ValueError("%02x%02x" % (sw[0], sw[1]))
        sw = self.icc_send_cmd(cmd_data1)
        if len(sw) != 2:
            raise ValueError(sw)
        elif sw[0] != 0x61:
            raise ValueError("%02x%02x" % (sw[0], sw[1]))
        return self.cmd_get_response(sw[1])

    def cmd_internal_authenticate(self, data):
        cmd_data = iso7816_compose(0x88, 0, 0, data)
        sw = self.icc_send_cmd(cmd_data)
        if len(sw) != 2:
            raise ValueError(sw)
        if sw[0] == 0x90 and sw[1] == 0x00:
            return array("B")
        elif sw[0] != 0x61:
            raise ValueError("%02x%02x" % (sw[0], sw[1]))
        return self.cmd_get_response(sw[1])

    def cmd_genkey(self, keyno):
        if keyno == 1:
            data = b"\xb6\x00"
        elif keyno == 2:
            data = b"\xb8\x00"
        else:
            data = b"\xa4\x00"
        cmd_data = iso7816_compose(0x47, 0x80, 0, data)
        sw = self.icc_send_cmd(cmd_data)
        if len(sw) != 2:
            raise ValueError(sw)
        if sw[0] == 0x90 and sw[1] == 0x00:
            return array("B")
        elif sw[0] != 0x61:
            raise ValueError("%02x%02x" % (sw[0], sw[1]))
        pk = self.cmd_get_response(sw[1])
        return (pk[9 : 9 + 256], pk[9 + 256 + 2 : 9 + 256 + 2 + 3])

    def cmd_get_public_key(self, keyno):
        if keyno == 1:
            data = b"\xb6\x00"
        elif keyno == 2:
            data = b"\xb8\x00"
        else:
            data = b"\xa4\x00"
        cmd_data = iso7816_compose(0x47, 0x81, 0, data)
        sw = self.icc_send_cmd(cmd_data)
        if len(sw) != 2:
            raise ValueError(sw)
        elif sw[0] != 0x61:
            raise ValueError("%02x%02x" % (sw[0], sw[1]))
        pk = self.cmd_get_response(sw[1])
        return (pk[9 : 9 + 256], pk[9 + 256 + 2 : 9 + 256 + 2 + 3])

    def cmd_put_data_remove(self, tagh, tagl):
        cmd_data = iso7816_compose(0xDA, tagh, tagl, b"")
        sw = self.icc_send_cmd(cmd_data)
        if sw[0] != 0x90 and sw[1] != 0x00:
            raise ValueError("%02x%02x" % (sw[0], sw[1]))

    def cmd_put_data_key_import_remove(self, keyno):
        if keyno == 1:
            keyspec = b"\xb6\x00"  # SIG
        elif keyno == 2:
            keyspec = b"\xb8\x00"  # DEC
        else:
            keyspec = b"\xa4\x00"  # AUT
        cmd_data = iso7816_compose(0xDB, 0x3F, 0xFF, b"\x4d\x02" + keyspec)
        sw = self.icc_send_cmd(cmd_data)
        if sw[0] != 0x90 and sw[1] != 0x00:
            raise ValueError("%02x%02x" % (sw[0], sw[1]))

    def cmd_get_challenge(self):
        cmd_data = iso7816_compose(0x84, 0x00, 0x00, "")
        sw = self.icc_send_cmd(cmd_data)
        if len(sw) != 2:
            raise ValueError(sw)
        if sw[0] != 0x61:
            raise ValueError("%02x%02x" % (sw[0], sw[1]))
        return self.cmd_get_response(sw[1])

    def cmd_external_authenticate(self, keyno, signed):
        cmd_data = iso7816_compose(0x82, 0x00, keyno, signed[0:128], cls=0x10)
        sw = self.icc_send_cmd(cmd_data)
        if len(sw) != 2:
            raise ValueError(sw)
        if not (sw[0] == 0x90 and sw[1] == 0x00):
            raise ValueError("%02x%02x" % (sw[0], sw[1]))
        cmd_data = iso7816_compose(0x82, 0x00, keyno, signed[128:])
        sw = self.icc_send_cmd(cmd_data)
        if len(sw) != 2:
            raise ValueError(sw)
        if not (sw[0] == 0x90 and sw[1] == 0x00):
            raise ValueError("%02x%02x" % (sw[0], sw[1]))


class regnual(object):
    def __init__(self, dev):
        conf = dev.configurations[0]
        intf_alt = conf.interfaces[0]
        intf = intf_alt[0]
        if intf.interfaceClass != 0xFF:
            raise ValueError("Wrong interface class")
        self.__devhandle = dev.open()
        self.__devhandle.claimInterface(intf)
        self.logger = logging.getLogger("regnual")

    def set_logger(self, logger: logging.Logger):
        self.logger = logger.getChild("regnual")

    def local_print(self, message: str, verbose=False):
        self.logger.debug("print: {}".format(message))
        if verbose:
            print(message)

    def mem_info(self):
        mem = self.__devhandle.controlMsg(
            requestType=0xC0, request=0, buffer=8, value=0, index=0, timeout=10000
        )
        start = ((mem[3] * 256 + mem[2]) * 256 + mem[1]) * 256 + mem[0]
        end = ((mem[7] * 256 + mem[6]) * 256 + mem[5]) * 256 + mem[4]
        return (start, end)

    def download(self, start, data, verbose=False, progress_func=None):
        addr = start
        addr_end = (start + len(data)) & 0xFFFFFF00
        i = int((addr - 0x08000000) / 0x100)
        j = 0
        self.local_print("start %08x" % addr, verbose)
        self.local_print("end   %08x" % addr_end, verbose)
        if progress_func:
            progress_func(0)
        while addr < addr_end:
            if progress_func:
                progress_func((addr - start) / (addr_end - start))
            self.local_print("# %08x: %d: %d : %d" % (addr, i, j, 256), verbose)
            self.__devhandle.controlMsg(
                requestType=0x40,
                request=1,
                buffer=data[j * 256 : j * 256 + 256],
                value=0,
                index=0,
                timeout=10000,
            )
            crc32code = crc32(data[j * 256 : j * 256 + 256])
            res = self.__devhandle.controlMsg(
                requestType=0xC0, request=2, buffer=4, value=0, index=0, timeout=10000
            )
            r_value = ((res[3] * 256 + res[2]) * 256 + res[1]) * 256 + res[0]
            if (crc32code ^ r_value) != 0xFFFFFFFF:
                self.local_print("failure")
            self.__devhandle.controlMsg(
                requestType=0x40,
                request=3,
                buffer=None,
                value=i,
                index=0,
                timeout=10000,
            )
            time.sleep(0.010)
            res = self.__devhandle.controlMsg(
                requestType=0xC0, request=2, buffer=4, value=0, index=0, timeout=10000
            )
            r_value = ((res[3] * 256 + res[2]) * 256 + res[1]) * 256 + res[0]
            if r_value == 0:
                self.local_print("failure")
            i = i + 1
            j = j + 1
            addr = addr + 256
        residue = len(data) % 256
        if residue != 0:
            self.local_print("# %08x: %d : %d" % (addr, i, residue), verbose)
            self.__devhandle.controlMsg(
                requestType=0x40,
                request=1,
                buffer=data[j * 256 :],
                value=0,
                index=0,
                timeout=10000,
            )
            crc32code = crc32(data[j * 256 :].ljust(256, b"\xff"))
            res = self.__devhandle.controlMsg(
                requestType=0xC0, request=2, buffer=4, value=0, index=0, timeout=10000
            )
            r_value = ((res[3] * 256 + res[2]) * 256 + res[1]) * 256 + res[0]
            if (crc32code ^ r_value) != 0xFFFFFFFF:
                self.local_print("failure")
            self.__devhandle.controlMsg(
                requestType=0x40,
                request=3,
                buffer=None,
                value=i,
                index=0,
                timeout=10000,
            )
            time.sleep(0.010)
            res = self.__devhandle.controlMsg(
                requestType=0xC0, request=2, buffer=4, value=0, index=0, timeout=10000
            )
            r_value = ((res[3] * 256 + res[2]) * 256 + res[1]) * 256 + res[0]
            if r_value == 0:
                self.local_print("failure")

    def protect(self):
        self.__devhandle.controlMsg(
            requestType=0x40, request=4, buffer=None, value=0, index=0, timeout=10000
        )
        time.sleep(0.100)
        res = self.__devhandle.controlMsg(
            requestType=0xC0, request=2, buffer=4, value=0, index=0, timeout=10000
        )
        r_value = ((res[3] * 256 + res[2]) * 256 + res[1]) * 256 + res[0]
        if r_value == 0:
            self.local_print("protection failure")

    def finish(self):
        self.__devhandle.controlMsg(
            requestType=0x40, request=5, buffer=None, value=0, index=0, timeout=10000
        )

    def reset_device(self):
        try:
            self.__devhandle.reset()
        except:
            pass


def compare(data_original, data_in_device):
    if data_original == data_in_device:
        return True
    raise ValueError("verify failed")


def gnuk_devices():
    busses = usb.busses()
    for bus in busses:
        devices = bus.devices
        for dev in devices:
            for config in dev.configurations:
                for intf in config.interfaces:
                    for alt in intf:
                        if (
                            alt.interfaceClass == CCID_CLASS
                            and alt.interfaceSubClass == CCID_SUBCLASS
                            and alt.interfaceProtocol == CCID_PROTOCOL_0
                            and (dev.idVendor, dev.idProduct) in USB_PRODUCT_LIST_TUP
                        ):
                            yield dev, config, alt


def gnuk_devices_by_vidpid():
    try:
        busses = usb.busses()
    except usb.core.NoBackendError:
        print(
            "Warning: no backend was found to use for communication. "
            "Please refer to documentation how to install additional libraries."
        )
        return []

    for bus in busses:
        devices = bus.devices
        for dev in devices:
            for cand in USB_PRODUCT_LIST:
                if dev.idVendor != cand["vendor"]:
                    continue
                if dev.idProduct != cand["product"]:
                    continue
                yield dev
                break


def get_gnuk_device(verbose=True, logger: Optional[logging.Logger] = None):
    from usb import USBError

    candidates = []
    icc = None
    for (dev, config, intf) in gnuk_devices():
        try:
            icc = gnuk_token(dev, config, intf)
            if logger:
                icc.set_logger(logger)
            if logger:
                logger.debug(
                    "{} {} {}".format(dev.filename, config.value, intf.interfaceNumber)
                )
            if verbose:
                try:
                    d = get_dict_for_device(dev)
                    print(f'Device: {d["Product"]} {d["Serial"]}')
                except:
                    print(
                        f'Device: name: "{dev.filename}", c/i: {config.value}/{intf.interfaceNumber}'
                    )
            break
        except USBError as e:
            # USBError(16, 'Resource busy')
            candidates.append(icc)
        except Exception:
            pass
    if not icc and candidates:
        raise OnlyBusyICCError(f"Found only busy ICC: {candidates}")
    if not icc:
        raise ValueError("No ICC present")
    status = icc.icc_get_status()
    if status == 0:
        pass  # It's ON already
    elif status == 1:
        icc.icc_power_on()
    else:
        raise ValueError("Unknown ICC status", status)
    return icc


SHA256_OID_PREFIX = "3031300d060960864801650304020105000420"


def UNSIGNED(n):
    return n & 0xFFFFFFFF


def crc32(bytestr):
    crc = binascii.crc32(bytestr)
    return UNSIGNED(crc)


def parse_kdf_data(kdf_data):
    if len(kdf_data) == 90:
        single_salt = True
    elif len(kdf_data) == 110:
        single_salt = False
    else:
        raise ValueError("length does not much", kdf_data)

    if kdf_data[0:2] != b"\x81\x01":
        raise ValueError("data does not much")
    algo = kdf_data[2]
    if kdf_data[3:5] != b"\x82\x01":
        raise ValueError("data does not much")
    subalgo = kdf_data[5]
    if kdf_data[6:8] != b"\x83\x04":
        raise ValueError("data does not much")
    iters = unpack(">I", kdf_data[8:12])[0]
    if kdf_data[12:14] != b"\x84\x08":
        raise ValueError("data does not much")
    salt = kdf_data[14:22]
    if single_salt:
        salt_reset = None
        salt_admin = None
        if kdf_data[22:24] != b"\x87\x20":
            raise ValueError("data does not much")
        hash_user = kdf_data[24:56]
        if kdf_data[56:58] != b"\x88\x20":
            raise ValueError("data does not much")
        hash_admin = kdf_data[58:90]
    else:
        if kdf_data[22:24] != b"\x85\x08":
            raise ValueError("data does not much")
        salt_reset = kdf_data[24:32]
        if kdf_data[32:34] != b"\x86\x08":
            raise ValueError("data does not much")
        salt_admin = kdf_data[34:42]
        if kdf_data[42:44] != b"\x87\x20":
            raise ValueError("data does not much")
        hash_user = kdf_data[44:76]
        if kdf_data[76:78] != b"\x88\x20":
            raise ValueError("data does not much")
        hash_admin = kdf_data[78:110]
    return (algo, subalgo, iters, salt, salt_reset, salt_admin, hash_user, hash_admin)


class OnlyBusyICCError(ValueError):
    pass



#
# Copyright (c) 2016 Nordic Semiconductor ASA
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright notice, this
#   list of conditions and the following disclaimer in the documentation and/or
#   other materials provided with the distribution.
#
#   3. Neither the name of Nordic Semiconductor ASA nor the names of other
#   contributors to this software may be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#   4. This software must only be used in or with a processor manufactured by Nordic
#   Semiconductor ASA, or in or with a processor manufactured by a third party that
#   is used in combination with a processor manufactured by Nordic Semiconductor.
#
#   5. Any software provided in binary or object form under this license must not be
#   reverse engineered, decompiled, modified and/or disassembled.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# Copyright (c) 2015 Nordic Semiconductor. All Rights Reserved.
#
# The information contained herein is property of Nordic Semiconductor ASA.
# Terms and conditions of usage are described in detail in NORDIC
# SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
#
# Licensees are granted free, non-transferable use of the information. NO
# WARRANTY of ANY KIND is provided. This heading must NOT be removed from
# the file.

import datetime
import hashlib

try:
    from ecdsa import SigningKey
    from ecdsa.curves import NIST256p
    from ecdsa.keys import sigencode_string
except Exception:
    print("Failed to import ecdsa, cannot do signing")


keys_default_pem = """-----BEGIN EC PRIVATE KEY-----
MHcCAQEEIGvsrpXh8m/E9bj1dq/0o1aBPQVAFJQ6Pzusx685URE0oAoGCCqGSM49
AwEHoUQDQgAEaHYrUu/oFKIXN457GH+8IOuv6OIPBRLqoHjaEKM0wIzJZ0lhfO/A
53hKGjKEjYT3VNTQ3Zq1YB3o5QSQMP/LRg==
-----END EC PRIVATE KEY-----"""


class Signing:
    """
    Class for singing of hex-files
    """

    def gen_key(self, filename):
        """
        Generate a new Signing key using NIST P-256 curve
        """
        self.sk = SigningKey.generate(curve=NIST256p)

        with open(filename, "wb") as sk_file:
            sk_file.write(self.sk.to_pem())

    def load_key(self, filename):
        """
        Load signing key (from pem file)
        """
        default_sk = SigningKey.from_pem(keys_default_pem)

        with open(filename, "r") as sk_file:
            sk_pem = sk_file.read()

        self.sk = SigningKey.from_pem(sk_pem)
        return default_sk.to_string() == self.sk.to_string()

    def sign(self, init_packet_data):
        """
        Create signature for init package using P-256 curve and SHA-256 as hashing algorithm
        Returns R and S keys combined in a 64 byte array
        """
        # Add assertion of init_packet
        if self.sk is None:
            raise AssertionError("Can't save key. No key created/loaded")

        # Sign the init-packet
        signature = self.sk.sign(
            init_packet_data, hashfunc=hashlib.sha256, sigencode=sigencode_string
        )
        return signature[31::-1] + signature[63:31:-1]

    def verify(self, init_packet, signature):
        """
        Verify init packet
        """
        # Add assertion of init_packet
        if self.sk is None:
            raise AssertionError("Can't save key. No key created/loaded")

        vk = self.sk.get_verifying_key()

        # Verify init packet
        try:
            vk.verify(signature, init_packet, hashfunc=hashlib.sha256)
        except:
            return False

        return True

    def get_vk(self, output_type, dbg) -> str:
        """
        Get public key (as hex, code or pem)
        """
        if self.sk is None:
            raise AssertionError("Can't get key. No key created/loaded")

        if output_type is None:
            raise ValueError("Invalid output type for public key.")
        elif output_type == "hex":
            return self.get_vk_hex()
        elif output_type == "code":
            return self.get_vk_code(dbg)
        elif output_type == "pem":
            return self.get_vk_pem()
        else:
            raise ValueError("Invalid argument. Can't get key")

    def get_sk(self, output_type, dbg) -> str:
        """
        Get private key (as hex, code or pem)
        """
        if self.sk is None:
            raise AssertionError("Can't get key. No key created/loaded")

        if output_type is None:
            raise ValueError("Invalid output type for private key.")
        elif output_type == "hex":
            return self.get_sk_hex()
        elif output_type == "code":
            raise ValueError("Private key cannot be shown as code")
        elif output_type == "pem":
            # Return pem as str to conform in type with the other cases.
            return self.sk.to_pem().decode()
        else:
            raise ValueError("Invalid argument. Can't get key")

    def get_sk_hex(self):
        """
        Get the verification key as hex
        """
        if self.sk is None:
            raise AssertionError("Can't get key. No key created/loaded")

        # Reverse the key for display. This emulates a memory
        # dump of the key interpreted a 256bit little endian
        # integer.
        key = self.sk.to_string()
        displayed_key = key[::-1].hex()

        return f"Private (signing) key sk:\n{displayed_key}"

    def get_vk_hex(self):
        """
        Get the verification key as hex
        """
        if self.sk is None:
            raise AssertionError("Can't get key. No key created/loaded")

        # Reverse the two halves of key for display. This
        # emulates a memory dump of the key interpreted as two
        # 256bit little endian integers.
        key = self.sk.get_verifying_key().to_string()
        displayed_key = (key[:32][::-1] + key[32:][::-1]).hex()

        return f"Public (verification) key pk:\n{displayed_key}"

    def wrap_code(self, key_code, dbg):

        header = """
/* This file was automatically generated by nrfutil on {0} */

#include "stdint.h"
#include "compiler_abstraction.h"
""".format(
            datetime.datetime.now().strftime("%Y-%m-%d (YY-MM-DD) at %H:%M:%S")
        )

        dbg_header = """
/* This file was generated with a throwaway private key, that is only intended for a debug version of the DFU project.
  Please see https://github.com/NordicSemiconductor/pc-nrfutil/blob/master/README.md to generate a valid public key. */

#ifdef NRF_DFU_DEBUG_VERSION
"""
        dbg_footer = """
#else
#error "Debug public key not valid for production. Please see https://github.com/NordicSemiconductor/pc-nrfutil/blob/master/README.md to generate it"
#endif
"""
        if dbg:
            code = header + dbg_header + key_code + dbg_footer
        else:
            code = header + key_code
        return code

    def get_vk_code(self, dbg):
        """
        Get the verification key as code
        """
        if self.sk is None:
            raise AssertionError("Can't get key. No key created/loaded")

        to_two_digit_hex_with_0x = "0x{0:02x}".format

        key = self.sk.get_verifying_key().to_string()
        vk_x_separated = ", ".join(map(to_two_digit_hex_with_0x, key[:32][::-1]))
        vk_y_separated = ", ".join(map(to_two_digit_hex_with_0x, key[32:][::-1]))

        key_code = """
/** @brief Public key used to verify DFU images */
__ALIGN(4) const uint8_t pk[64] =
{{
    {0},
    {1}
}};
"""
        key_code = key_code.format(vk_x_separated, vk_y_separated)
        vk_code = self.wrap_code(key_code, dbg)

        return vk_code

    def get_vk_pem(self) -> str:
        """
        Get the verification key as PEM
        """
        if self.sk is None:
            raise AssertionError("Can't get key. No key created/loaded")

        vk = self.sk.get_verifying_key()
        vk_pem = vk.to_pem()

        # Return pem as str to conform in type with the other cases.
        return vk_pem.decode()



# -*- coding: utf-8 -*-
#
# Copyright 2020 Nitrokey Developers
#
# Licensed under the Apache License, Version 2.0, <LICENSE-APACHE or
# http://apache.org/licenses/LICENSE-2.0> or the MIT license <LICENSE-MIT or
# http://opensource.org/licenses/MIT>, at your option. This file may not be
# copied, modified, or distributed except according to those terms.
import logging
import platform
import string
import subprocess
import time
from shutil import which
from typing import Optional

import click
import usb1
from intelhex import IntelHex
from tqdm import tqdm

from pynitrokey.cli.exceptions import CliException
from pynitrokey.helpers import (
    AskUser,
    check_pynitrokey_version,
    confirm,
    local_critical,
    local_print,
    prompt,
)
from pynitrokey.libnk import DeviceNotFound, NitrokeyStorage, RetCode


def connect_nkstorage():
    try:
        nks = NitrokeyStorage()
        nks.connect()
        return nks
    except DeviceNotFound:
        raise CliException("No Nitrokey Storage device found", support_hint=False)


logger = logging.getLogger(__name__)


@click.group()
def storage():
    """Interact with Nitrokey Storage devices, see subcommands."""
    pass


def process_runner(c: str, args: Optional[dict] = None) -> str:
    """Wrapper for running command and returning output, both logged"""
    cmd = c.split()
    if args and any(f"${key}" in c for key in args.keys()):
        for i, _ in enumerate(cmd):
            template = string.Template(cmd[i])
            cmd[i] = template.substitute(args)

    logger.debug(f"Running {c}")
    local_print(f'* Running \t"{c}"')
    try:
        output = subprocess.check_output(cmd, stderr=subprocess.STDOUT).decode()
    except subprocess.CalledProcessError as e:
        logger.error(f'Output for "{c}": {e.output}')
        local_print(f'\tOutput for "{c}": "{e.output.strip().decode()}"')
        raise
    logger.debug(f'Output for "{c}": {output}')
    return output


class DfuTool:
    name = "dfu-programmer"

    @classmethod
    def is_available(cls):
        """Check whether `name` is on PATH and marked as executable."""
        return which(cls.name) is not None

    @classmethod
    def get_version(cls) -> str:
        c = f"{cls.name} --version"
        output = process_runner(c).strip()
        return output

    @classmethod
    def check_version(cls) -> bool:
        # todo choose and use specialized package for version strings management, e.g:
        #   from packaging import version
        ver_string = cls.get_version()
        ver = ver_string.split()[1]
        ver_found = (*map(int, ver.split(".")),)
        ver_required = (0, 6, 1)
        local_print(f"Tool found: {ver_string}")
        return ver_found >= ver_required

    @classmethod
    def self_check(cls) -> bool:
        if not cls.is_available():
            local_print(
                f"{cls.name} is not available. Please install it or use another tool for update."
            )
            raise click.Abort()

        local_print("")
        cls.check_version()
        local_print("")
        return True


class ConnectedDevices:
    application_mode: int
    update_mode: int

    def __init__(self, application_mode, update_mode):
        self.application_mode = application_mode
        self.update_mode = update_mode


class UsbId:
    vid: int
    pid: int

    def __init__(self, vid, pid):
        self.vid = vid
        self.pid = pid


def is_connected() -> ConnectedDevices:
    devs = {}
    usb_id = {
        "update_mode": UsbId(0x03EB, 0x2FF1),
        "application_mode": UsbId(0x20A0, 0x4109),
    }
    with usb1.USBContext() as context:
        for k, v in usb_id.items():
            res = context.getByVendorIDAndProductID(vendor_id=v.vid, product_id=v.pid)
            devs[k] = 1 if res else 0
    return ConnectedDevices(
        application_mode=devs["application_mode"], update_mode=devs["update_mode"]
    )


@click.command()
@click.argument("firmware", type=click.Path(exists=True, readable=True))
@click.option(
    "--experimental",
    default=False,
    is_flag=True,
    help="Allow to execute experimental features",
)
def update(firmware: str, experimental):
    """experimental: run assisted update through dfu-programmer tool"""
    check_pynitrokey_version()

    if platform.system() != "Linux" or not experimental:
        local_print(
            "This feature is Linux only and experimental, which means it was not tested thoroughly.\n"
            "Please pass --experimental switch to force running it anyway."
        )
        raise click.Abort()
    assert firmware.endswith(".hex")

    DfuTool.self_check()

    commands = """
        dfu-programmer at32uc3a3256s erase
        dfu-programmer at32uc3a3256s flash --suppress-bootloader-mem $FIRMWARE
        dfu-programmer at32uc3a3256s start
        """

    local_print(
        "Note: During the execution update program will try to connect to the device. "
        "Check your udev rules in case of connection issues."
    )
    local_print(f"Using firmware path: {firmware}")
    # note: this is just for presentation - actual argument replacement is done in process_runner
    # the string form cannot be used, as it could contain space which would break dfu-programmer's call
    args = {"FIRMWARE": firmware}
    local_print(
        f"Commands to be executed: {string.Template(commands).substitute(args)}"
    )
    if not confirm("Do you want to perform the firmware update now?"):
        logger.info("Update cancelled by user")
        raise click.Abort()

    check_for_update_mode()

    commands_clean = commands.strip().split("\n")
    for c in commands_clean:
        c = c.strip()
        if not c:
            continue
        try:
            output = process_runner(c, args)
            if output:
                local_print(output)
        except subprocess.CalledProcessError as e:
            linux = "linux" in platform.platform().lower()
            local_critical(
                e, "Note: make sure you have the udev rules installed." if linux else ""
            )

    local_print("")
    local_print("Finished!")

    for _ in tqdm(range(10), leave=False):
        if is_connected().application_mode != 0:
            break
        time.sleep(1)

    list_cmd = storage.commands["list"]
    assert list_cmd.callback
    list_cmd.callback()


def check_for_update_mode():
    connected = is_connected()
    assert (
        connected.application_mode + connected.update_mode > 0
    ), "No connected Nitrokey Storage devices found"
    if connected.application_mode and not connected.update_mode:
        # execute bootloader
        storage.commands["enable-update"].callback()
        for _ in tqdm(range(10), leave=False):
            if is_connected().update_mode != 0:
                break
            time.sleep(1)
        time.sleep(1)
    else:
        local_print(
            "Nitrokey Storage in update mode found in the USB list (not connected yet)"
        )


@click.command()
def list():
    """List connected devices"""

    local_print(":: 'Nitrokey Storage' keys:")
    devices = NitrokeyStorage.list_devices()
    for dct in devices:
        local_print(f" - {dct}")
    if len(devices) == 1:
        nks = NitrokeyStorage()
        nks.connect()
        local_print(f"Found libnitrokey version: {nks.library_version()}")
        local_print(f"Firmware version: {nks.fw_version}")
        local_print(f"Admin PIN retries: {nks.admin_pin_retries}")
        local_print(f"User PIN retries: {nks.user_pin_retries}")


@click.command()
def enable_update():
    """Enable firmware update for NK Storage device

    If the Firmware Password is not in the environment variable NITROPY_FIRMWARE_PASSWORD, it will be prompted from stdin
    """
    password = AskUser(
        "Firmware Password", envvar="NITROPY_FIRMWARE_PASSWORD", hide_input=True
    ).ask()
    local_print("Enabling firmware update mode")
    nks = connect_nkstorage()
    if nks.enable_firmware_update(password) == 0:
        local_print("setting firmware update mode - success!")
    else:
        local_critical(
            "Enabling firmware update has failed. Check your firmware password."
        )


@click.command()
def change_firmware_password():
    """
    Change the firmware update password.

    The user is prompted for the old and the new firmware update password.  Per
    default, the firmware update password is 12345678.
    """
    nk = connect_nkstorage()

    old_password = prompt(
        "Old firmware update password", default="12345678", hide_input=True
    )
    new_password = prompt(
        "New firmware update password", hide_input=True, confirmation_prompt=True
    )
    ret = nk.change_firmware_password(old_password, new_password)
    if ret.ok:
        local_print("Successfully updated the firmware password")
    elif ret == RetCode.WRONG_PASSWORD:
        local_critical("Wrong firmware update password", support_hint=False)
    elif ret == RetCode.TooLongStringException:
        local_critical(
            "The new firmware update password is too long", support_hint=False
        )
    else:
        local_critical(f"Failed to update the firmware password ({ret.name})")


@click.command()
def open_encrypted():
    """Unlock the encrypted volume

    If the User PIN is not in the environment variable NITROPY_USER_PIN, it will be prompted from stdin
    """
    password = AskUser("User PIN", envvar="NITROPY_USER_PIN", hide_input=True).ask()
    nks = connect_nkstorage()
    ret = nks.unlock_encrypted_volume(password)
    if not ret.ok:
        if ret == RetCode.WRONG_PASSWORD:
            raise CliException("Wrong user PIN", support_hint=False)
        else:
            raise CliException(
                "Unexpected error unlocking the encrypted volume {}".format(str(ret))
            )


@click.command()
def close_encrypted():
    """Lock the encrypted volume"""
    nks = connect_nkstorage()
    ret = nks.lock_encrypted_volume()
    if not ret.ok:
        raise CliException("Error closing the encrypted volume: {}".format(str(ret)))


@click.command()
def open_hidden():
    """Unlock a hidden volume

    If the hidden volume passphrase is not in the environment variable NITROPY_HIDDEN_PASSPHRASE, it will be prompted from stdin
    """
    password = AskUser(
        "Hidden volume passphrase", envvar="NITROPY_HIDDEN_PASSPHRASE", hide_input=True
    ).ask()
    nks = connect_nkstorage()
    ret = nks.unlock_hidden_volume(password)
    if not ret.ok:
        if ret == RetCode.WRONG_PASSWORD:
            raise CliException("Wrong hidden volume passphrase", support_hint=False)
        else:
            raise CliException(
                "Unexpected error unlocking the hidden volume: {}".format(str(ret))
            )


@click.command()
def close_hidden():
    """Lock the hidden volumes"""
    nks = connect_nkstorage()
    ret = nks.lock_hidden_volume()
    if not ret.ok:
        raise CliException("Error closing the hidden volume: {}".format(str(ret)))


@click.command()
@click.argument(
    "slot",
    type=int,
)
@click.argument(
    "begin",
    type=int,
)
@click.argument("end", type=int)
def create_hidden(slot, begin, end):
    """Create a hidden volume

    SLOT is the slot used for the hidden volume (1-4)\n
    START is where the volume begins expressed in percent of total available storage (0-99)\n
    END is where the volume ends expressed in percent of total available storage (1-100)\n
    If the hidden volume passphrase is not in the environment variable NITROPY_HIDDEN_PASSPHRASE, it will be prompted from stdin
    """
    if not slot in [1, 2, 3, 4]:
        raise CliException("Error: Slot must be between 1 and 4", support_hint=False)
    elif begin > 99 or begin < 0:
        raise CliException("Error: Begin must be between 0 and 99", support_hint=False)
    elif end < 1 or end > 100:
        raise CliException("Error: End must be between 1 and 100", support_hint=False)
    elif begin >= end:
        raise CliException(
            "Error: END must be strictly superior than START", support_hint=False
        )

    password = AskUser(
        "Hidden volume passphrase", envvar="NITROPY_HIDDEN_PASSPHRASE", hide_input=True
    ).ask()

    nks = connect_nkstorage()
    ret = nks.create_hidden_volume(slot - 1, begin, end, password)
    if not ret.ok:
        raise CliException("Error creating the hidden volume: {}".format(str(ret)))


class MemoryConstants:
    HEX_OFFSET = 0x80000000
    APPLICATION_DATA_START = 0x2000
    USER_DATA_START = 495 * 512  # 0x3DE00
    # user data end, last page is for the bootloader data (as per dfu-programmer manual)
    USER_DATA_END = 511 * 512  # 0x3FE00


def input_format(x: str):
    return "hex" if x.endswith("hex") else "bin"


def empty_check_user_data(ih: IntelHex):
    empty = 0
    for i in range(MemoryConstants.USER_DATA_START, MemoryConstants.USER_DATA_END):
        empty += ih[i] not in [0xFF, 0x00]
    return empty == 0


@click.command()
@click.argument("firmware", type=click.Path(exists=True))
def check(firmware: str):
    """Check if provided binary image contains user data in the proper region

    Use it on downloaded full image with `--force` flag, as in: \n
    $ dfu-programmer at32uc3a3256s read --bin --force > dump.bin
    """
    current_firmware_read = IntelHex()
    current_firmware_read.loadfile(firmware, format=input_format(firmware))

    if empty_check_user_data(current_firmware_read):
        raise click.ClickException(
            f"{firmware}: Provided dumped binary image does not contain user data"
        )
    click.echo(f"{firmware}: User data seem to be present")


@click.command()
@click.argument("fw1_path", type=click.Path(exists=True))
@click.argument("fw2_path", type=click.Path(exists=True))
@click.argument("region", type=click.Choice(["application", "user"]))
@click.option("--max-diff", type=int, default=10)
def compare(fw1_path: str, fw2_path: str, region: str, max_diff: int):
    """Compare two binary images"""

    fw1 = IntelHex()
    fw1.loadfile(fw1_path, format=input_format(fw1_path))
    fw2 = IntelHex()
    fw2.loadfile(fw2_path, format=input_format(fw2_path))

    offset = {}
    for f in [fw1, fw2]:
        offset[f] = 0
        if f.minaddr() >= MemoryConstants.HEX_OFFSET:
            offset[f] = MemoryConstants.HEX_OFFSET

    if fw1.minaddr() != fw2.minaddr():
        click.echo(
            f"Warning: different offsets found - this could make the operation fail: {hex(fw1.minaddr())} {hex(fw2.minaddr())}"
        )

    diff_count = 0
    non_empty_count = 0
    if region == "application":
        data_start = MemoryConstants.APPLICATION_DATA_START
        data_stop = MemoryConstants.USER_DATA_START
    elif region == "user":
        data_start = MemoryConstants.USER_DATA_START
        data_stop = MemoryConstants.USER_DATA_END
    else:
        raise click.ClickException(f"Wrong type")

    def geti(f, i):
        return f[i + offset[f]]

    click.echo(f"Checking binary images in range {hex(data_start)}:{hex(data_stop)}")
    for i in range(data_start, data_stop):
        fw1_i = geti(fw1, i)
        fw2_i = geti(fw2, i)
        data_equal = fw1_i == fw2_i or fw1_i in [0xFF, 0x00] and fw2_i in [0xFF, 0x00]
        diff_count += not data_equal
        non_empty_count += fw1_i not in [0xFF, 0x00]
        if not data_equal:
            click.echo(
                f"Binaries differ at {hex(i)} (page {i // 512}): {hex(fw1_i)} {hex(fw2_i)}"
            )
        if diff_count > max_diff:
            raise click.ClickException(f"Maximum diff count reached")

    if diff_count > 0:
        raise click.ClickException(f"Binaries differ")
    if non_empty_count == 0:
        raise click.ClickException(f"Binaries contain no data")
    click.echo(f"Non-empty bytes count: {non_empty_count}")
    click.echo("Binary images are identical")


@click.command()
@click.argument("dumped_firmware", type=click.Path(exists=True))
@click.argument("new_firmware_file", type=click.Path(exists=True))
@click.argument("output", type=click.File("w"))
@click.option("--overlap", type=click.Choice(["error", "ignore"]), default="error")
def merge(
    dumped_firmware: str,
    new_firmware_file: str,
    output: click.File,
    overlap: str,
):
    """Simple tool to merge user data into the new firmware binary"""
    if not output.name.endswith("hex"):
        raise click.ClickException("Provided output path has to end in .hex")

    current_firmware_read = IntelHex()
    current_firmware_read.loadfile(
        dumped_firmware, format=input_format(dumped_firmware)
    )

    if empty_check_user_data(current_firmware_read):
        raise click.ClickException(
            "Provided dumped binary image does not contain user data"
        )

    new_firmware = IntelHex()
    new_firmware.loadfile(new_firmware_file, format=input_format(new_firmware_file))
    new_firmware.merge(
        current_firmware_read[
            MemoryConstants.USER_DATA_START : MemoryConstants.USER_DATA_END
        ],
        overlap=overlap,
    )
    new_firmware.write_hex_file(output)
    click.echo(f'Done. Results written to "{output.name}".')


@click.group()
def user_data():
    """experimental: commands to check and manipulate user data in the downloaded binary images"""
    pass


user_data.add_command(merge)
user_data.add_command(check)
user_data.add_command(compare)

storage.add_command(list)
storage.add_command(enable_update)
storage.add_command(change_firmware_password)
storage.add_command(open_encrypted)
storage.add_command(close_encrypted)
storage.add_command(open_hidden)
storage.add_command(close_hidden)
storage.add_command(create_hidden)
storage.add_command(update)
storage.add_command(user_data)



# -*- coding: utf-8 -*-
#
# Copyright 2021-2022 Nitrokey Developers
#
# Licensed under the Apache License, Version 2.0, <LICENSE-APACHE or
# http://apache.org/licenses/LICENSE-2.0> or the MIT license <LICENSE-MIT or
# http://opensource.org/licenses/MIT>, at your option. This file may not be
# copied, modified, or distributed except according to those terms.

import logging
import os.path
import sys
from hashlib import sha256
from typing import BinaryIO, Callable, List, Optional, Type, TypeVar

import click
from cryptography import x509
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePublicKey
from ecdsa import NIST256p, SigningKey

from pynitrokey.cli.exceptions import CliException
from pynitrokey.helpers import (
    DownloadProgressBar,
    Retries,
    check_experimental_flag,
    local_print,
    require_windows_admin,
)
from pynitrokey.nk3 import list as list_nk3
from pynitrokey.nk3 import open as open_nk3
from pynitrokey.nk3.admin_app import AdminApp
from pynitrokey.nk3.base import Nitrokey3Base
from pynitrokey.nk3.bootloader import (
    FirmwareContainer,
    Nitrokey3Bootloader,
    parse_firmware_image,
)
from pynitrokey.nk3.device import BootMode, Nitrokey3Device
from pynitrokey.nk3.exceptions import TimeoutException
from pynitrokey.nk3.provisioner_app import ProvisionerApp
from pynitrokey.nk3.updates import REPOSITORY, get_firmware_update
from pynitrokey.updates import OverwriteError

T = TypeVar("T", bound=Nitrokey3Base)

logger = logging.getLogger(__name__)


class Context:
    def __init__(self, path: Optional[str]) -> None:
        self.path = path

    def list(self) -> List[Nitrokey3Base]:
        if self.path:
            device = open_nk3(self.path)
            if device:
                return [device]
            else:
                return []
        else:
            return list_nk3()

    def _select_unique(self, name: str, devices: List[T]) -> T:
        if len(devices) == 0:
            msg = f"No {name} device found"
            if self.path:
                msg += f" at path {self.path}"
            raise CliException(msg)

        if len(devices) > 1:
            raise CliException(
                f"Multiple {name} devices found -- use the --path option to select one"
            )

        return devices[0]

    def connect(self) -> Nitrokey3Base:
        return self._select_unique("Nitrokey 3", self.list())

    def connect_device(self) -> Nitrokey3Device:
        devices = [
            device for device in self.list() if isinstance(device, Nitrokey3Device)
        ]
        return self._select_unique("Nitrokey 3", devices)

    def _await(
        self,
        name: str,
        ty: Type[T],
        retries: int,
        callback: Optional[Callable[[int, int], None]] = None,
    ) -> T:
        for t in Retries(retries):
            logger.debug(f"Searching {name} device ({t})")
            devices = [device for device in self.list() if isinstance(device, ty)]
            if len(devices) == 0:
                if callback:
                    callback(int((t.i / retries) * 100), 100)
                logger.debug(f"No {name} device found, continuing")
                continue
            if len(devices) > 1:
                raise CliException(f"Multiple {name} devices found")
            if callback:
                callback(100, 100)
            return devices[0]

        raise CliException(f"No {name} device found")

    def await_device(
        self,
        retries: Optional[int] = 30,
        callback: Optional[Callable[[int, int], None]] = None,
    ) -> Nitrokey3Device:
        assert isinstance(retries, int)
        return self._await("Nitrokey 3", Nitrokey3Device, retries, callback)

    def await_bootloader(
        self,
        retries: Optional[int] = 30,
        callback: Optional[Callable[[int, int], None]] = None,
    ) -> Nitrokey3Bootloader:
        assert isinstance(retries, int)
        # mypy does not allow abstract types here, but this is still valid
        return self._await("Nitrokey 3 bootloader", Nitrokey3Bootloader, retries, callback)  # type: ignore


@click.group()
@click.option("-p", "--path", "path", help="The path of the Nitrokey 3 device")
@click.pass_context
def nk3(ctx: click.Context, path: Optional[str]) -> None:
    """Interact with Nitrokey 3 devices, see subcommands."""
    ctx.obj = Context(path)
    require_windows_admin()


@nk3.command()
def list() -> None:
    """List all Nitrokey 3 devices."""
    local_print(":: 'Nitrokey 3' keys")
    for device in list_nk3():
        with device as device:
            uuid = device.uuid()
            if uuid:
                local_print(f"{device.path}: {device.name} {uuid}")
            else:
                local_print(f"{device.path}: {device.name}")


@nk3.command()
@click.option(
    "--bootloader",
    is_flag=True,
    help="Reboot a Nitrokey 3 device into bootloader mode",
)
@click.pass_obj
def reboot(ctx: Context, bootloader: bool) -> None:
    """
    Reboot the key.

    Per default, the key will reboot into regular firmware mode.  If the --bootloader option
    is set, a key can boot from firmware mode to bootloader mode.  Booting into
    bootloader mode has to be confirmed by pressing the touch button.
    """
    with ctx.connect() as device:
        if bootloader:
            if isinstance(device, Nitrokey3Device):
                success = reboot_to_bootloader(device)
            else:
                raise CliException(
                    "A Nitrokey 3 device in bootloader mode can only reboot into firmware mode.",
                    support_hint=False,
                )
        else:
            success = device.reboot()

    if not success:
        raise CliException(
            "The connected device cannot be rebooted automatically.  Remove and reinsert the "
            "device to reboot it.",
            support_hint=False,
        )


def reboot_to_bootloader(device: Nitrokey3Device) -> bool:
    local_print(
        "Please press the touch button to reboot the device into bootloader mode ..."
    )
    try:
        return device.reboot(BootMode.BOOTROM)
    except TimeoutException:
        raise CliException(
            "The reboot was not confirmed with the touch button.",
            support_hint=False,
        )


@nk3.command()
@click.option(
    "-l",
    "--length",
    "length",
    default=57,
    help="The length of the generated data (default: 57)",
)
@click.pass_obj
def rng(ctx: Context, length: int) -> None:
    """Generate random data on the device."""
    with ctx.connect_device() as device:
        while length > 0:
            rng = device.rng()
            local_print(rng[:length].hex())
            length -= len(rng)


@nk3.command()
@click.option(
    "--pin",
    "pin",
    help="The FIDO2 PIN of the device (if enabled)",
)
@click.option(
    "--only",
    "only",
    help="Run only the specified tests (may not be used with --all, --include or --exclude)",
)
@click.option(
    "--all",
    "all",
    is_flag=True,
    default=False,
    help="Run all tests (except those specified with --exclude)",
)
@click.option(
    "--include",
    "include",
    help="Also run the specified tests",
)
@click.option(
    "--exclude",
    "exclude",
    help="Do not run the specified tests",
)
@click.option(
    "--list",
    "list_",
    is_flag=True,
    default=False,
    help="List the selected tests instead of running them",
)
@click.pass_obj
def test(
    ctx: Context,
    pin: Optional[str],
    only: Optional[str],
    all: bool,
    include: Optional[str],
    exclude: Optional[str],
    list_: bool,
) -> None:
    """Run some tests on all connected Nitrokey 3 devices."""
    from .test import (
        TestContext,
        TestSelector,
        list_tests,
        log_devices,
        log_system,
        run_tests,
    )

    test_selector = TestSelector(all=all)
    if only:
        if all or include or exclude:
            raise CliException(
                "--only may not be used together with --all, --include or --exclude.",
                support_hint=False,
            )
        test_selector.only = only.split(",")
    if include:
        test_selector.include = include.split(",")
    if exclude:
        test_selector.exclude = exclude.split(",")

    if list_:
        list_tests(test_selector)
        return

    log_system()
    devices = ctx.list()

    if len(devices) == 0:
        log_devices()
        raise CliException("No connected Nitrokey 3 devices found")

    local_print(f"Found {len(devices)} Nitrokey 3 device(s):")
    for device in devices:
        local_print(f"- {device.name} at {device.path}")

    results = []
    test_ctx = TestContext(pin=pin)
    for device in devices:
        results.append(run_tests(test_ctx, device, test_selector))

    n = len(devices)
    success = sum(results)
    failure = n - success
    local_print("")
    local_print(
        f"Summary: {n} device(s) tested, {success} successful, {failure} failed"
    )

    if failure > 0:
        local_print("")
        raise CliException(f"Test failed for {failure} device(s)")


@nk3.command()
@click.argument("path", default=".")
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Overwrite the firmware image if it already exists",
)
@click.option("--version", help="Download this version instead of the latest one")
def fetch_update(path: str, force: bool, version: Optional[str]) -> None:
    """
    Fetches a firmware update for the Nitrokey 3 and stores it at the given path.

    If no path is given, the firmware image stored in the current working
    directory.  If the given path is a directory, the image is stored under
    that directory.  Otherwise it is written to the path.  Existing files are
    only overwritten if --force is set.

    Per default, the latest firmware release is fetched.  If you want to
    download a specific version, use the --version option.
    """
    try:
        release = REPOSITORY.get_release_or_latest(version)
        update = get_firmware_update(release)
    except Exception as e:
        if version:
            raise CliException(f"Failed to find firmware update {version}", e)
        else:
            raise CliException("Failed to find latest firmware update", e)

    bar = DownloadProgressBar(desc=update.tag)

    try:
        if os.path.isdir(path):
            path = update.download_to_dir(path, overwrite=force, callback=bar.update)
        else:
            if not force and os.path.exists(path):
                raise OverwriteError(path)
            with open(path, "wb") as f:
                update.download(f, callback=bar.update)

        bar.close()

        local_print(f"Successfully downloaded firmware release {update.tag} to {path}")
    except OverwriteError as e:
        raise CliException(
            f"{e.path} already exists.  Use --force to overwrite the file.",
            support_hint=False,
        )
    except Exception as e:
        raise CliException(f"Failed to download firmware update {update.tag}", e)


@nk3.command()
@click.argument("image", type=click.Path(exists=True, dir_okay=False))
def validate_update(image: str) -> None:
    """
    Validates the given firmware image and prints the firmware version and the signer for all
    available variants.
    """
    container = FirmwareContainer.parse(image)
    print(f"version:      {container.version}")
    if container.pynitrokey:
        print(f"pynitrokey:   >= {container.pynitrokey}")

    for variant in container.images:
        data = container.images[variant]
        try:
            metadata = parse_firmware_image(variant, data)
        except Exception as e:
            raise CliException("Failed to parse and validate firmware image", e)

        signed_by = metadata.signed_by or "unsigned"

        print(f"variant:      {variant.value}")
        print(f"  version:    {metadata.version}")
        print(f"  signed by:  {signed_by}")

        if container.version != metadata.version:
            raise CliException(
                f"The firmware image for the {variant} variant and the release {version} has an "
                f"unexpected product version ({metadata.version})."
            )


@nk3.command()
@click.argument("image", type=click.Path(exists=True, dir_okay=False), required=False)
@click.option(
    "--version",
    help="Set the firmware version to update to (default: latest stable)",
)
@click.option(
    "--ignore-pynitrokey-version",
    default=False,
    is_flag=True,
    help="Allow updates with an outdated pynitrokey version (dangerous)",
)
@click.option(
    "--experimental",
    default=False,
    is_flag=True,
    help="Allow to execute experimental features",
    hidden=True,
)
@click.pass_obj
def update(
    ctx: Context,
    image: Optional[str],
    version: Optional[str],
    ignore_pynitrokey_version: bool,
    experimental: bool,
) -> None:
    """
    Update the firmware of the device using the given image.

    This command requires that exactly one Nitrokey 3 in bootloader or firmware mode is connected.
    The user is asked to confirm the operation before the update is started.  The Nitrokey 3 may
    not be removed during the update.  Also, additional Nitrokey 3 devices may not be connected
    during the update.

    If no firmware image is given, the latest firmware release is downloaded automatically.  If
    the --version option is set, the given version is downloaded instead.

    If the connected Nitrokey 3 device is in firmware mode, the user is prompted to touch the
    device’s button to confirm rebooting to bootloader mode.
    """

    if experimental:
        "The --experimental switch is not required to run this command anymore and can be safely removed."

    from .update import update as exec_update

    exec_update(ctx, image, version, ignore_pynitrokey_version)


@nk3.command()
@click.pass_obj
def status(ctx: Context) -> None:
    """Query the device status."""
    with ctx.connect_device() as device:
        uuid = device.uuid()
        if uuid is not None:
            local_print(f"UUID:               {uuid}")

        admin = AdminApp(device)
        version = admin.version()
        local_print(f"Firmware version:   {version}")

        status = admin.status()
        if status.init_status is not None:
            local_print(f"Init status:        {status.init_status}")
        if status.ifs_blocks is not None:
            local_print(f"Free blocks (int):  {status.ifs_blocks}")
        if status.efs_blocks is not None:
            local_print(f"Free blocks (ext):  {status.efs_blocks}")
        if status.variant is not None:
            local_print(f"Variant:            {status.variant.name}")


@nk3.command()
@click.pass_obj
@click.argument("key")
def get_config(ctx: Context, key: str) -> None:
    """Query a config value."""
    with ctx.connect_device() as device:
        admin = AdminApp(device)
        value = admin.get_config(key)
        print(value)


@nk3.command()
@click.pass_obj
@click.argument("key")
@click.argument("value")
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Set the config value even if it is not known to pynitrokey",
)
@click.option(
    "--dry-run",
    is_flag=True,
    default=False,
    help="Perform all checks but don’t execute the configuration change",
)
def set_config(ctx: Context, key: str, value: str, force: bool, dry_run: bool) -> None:
    """
    Set a config value.

    Per default, this command can only be used with configuration values that
    are known to pynitrokey.  Changing some configuration values can have side
    effects.  For these values, a summary of the effects of the change and a
    confirmation prompt will be printed.

    If you use the --force/-f flag, you can also set configuration values that
    are not known to pynitrokey.  This may have unexpected side effects, for
    example resetting an application.  It is only intended for development and
    testing.

    To see the information about a config value without actually performing the
    change, use the --dry-run flag.
    """

    with ctx.connect_device() as device:
        admin = AdminApp(device)

        # before the confirmation prompt, check if the config value is supported
        if not admin.has_config(key):
            raise CliException(
                f"The configuration option '{key}' is not supported by the device.",
                support_hint=False,
            )

        # config fields that don’t have side effects
        whitelist = [
            "fido.disable_skip_up_timeout",
        ]
        requires_touch = False
        requires_reboot = False

        if key == "opcard.use_se050_backend":
            requires_touch = True
            requires_reboot = True
            print(
                "This configuration values determines whether the OpenPGP Card "
                "application uses a software implementation or the secure element.",
                file=sys.stderr,
            )
            print(
                "Changing this configuration value will cause a factory reset of "
                "the OpenPGP card application and destroy all OpenPGP keys and "
                "user data currently stored on the device.",
                file=sys.stderr,
            )
        elif key not in whitelist:
            pass
            print(
                "Changing configuration values can have unexpected side effects, including data loss.",
                file=sys.stderr,
            )
            print(
                "This should only be used for development and testing.",
                file=sys.stderr,
            )

            if not force:
                raise CliException(
                    "Unknown config values can only be set if the --force/-f flag is set.  Aborting.",
                    support_hint=False,
                )

        if key not in whitelist:
            click.confirm("Do you want to continue anyway?", abort=True)

        if dry_run:
            print("Stopping dry run.", file=sys.stderr)
            raise click.Abort()

        if requires_touch:
            print(
                "Press the touch button to confirm the configuration change.",
                file=sys.stderr,
            )

        admin.set_config(key, value)

        if requires_reboot:
            print("Rebooting device to apply config change.")
            device.reboot()

        print(f"Updated configuration {key}.")


@nk3.command()
@click.pass_obj
def version(ctx: Context) -> None:
    """Query the firmware version of the device."""
    with ctx.connect_device() as device:
        version = device.version()
        local_print(version)


@nk3.command()
@click.pass_obj
@click.option(
    "--experimental",
    default=False,
    is_flag=True,
    help="Allow to execute experimental features",
    hidden=True,
)
def factory_reset(ctx: Context, experimental: bool) -> None:
    """Factory reset all functionality of the device"""
    check_experimental_flag(experimental)
    with ctx.connect_device() as device:
        device.factory_reset()


# We consciously do not allow resetting the admin app
APPLICATIONS_CHOICE = click.Choice(["fido", "opcard", "secrets", "piv", "webcrypt"])


@nk3.command()
@click.pass_obj
@click.argument("application", type=APPLICATIONS_CHOICE, required=True)
@click.option(
    "--experimental",
    default=False,
    is_flag=True,
    help="Allow to execute experimental features",
    hidden=True,
)
def factory_reset_app(ctx: Context, application: str, experimental: bool) -> None:
    """Factory reset all functionality of an application"""
    check_experimental_flag(experimental)
    with ctx.connect_device() as device:
        device.factory_reset_app(application)


@nk3.command()
@click.pass_obj
def wink(ctx: Context) -> None:
    """Send wink command to the device (blinks LED a few times)."""
    with ctx.connect_device() as device:
        device.wink()


@nk3.group(hidden=True)
def provision() -> None:
    """
    Provision the device.  This command is only used during the production
    process and not available for regular devices.
    """
    pass


@provision.command("fido2")
@click.pass_obj
@click.option(
    "--key",
    "key_file",
    required=True,
    type=click.File("rb"),
    help="The path of the FIDO2 attestation key",
)
@click.option(
    "--cert",
    "cert_file",
    required=True,
    type=click.File("rb"),
    help="The path of the FIDO2 attestation certificate",
)
def provision_fido2(ctx: Context, key_file: BinaryIO, cert_file: BinaryIO) -> None:
    """Provision the FIDO2 attestation key and certificate."""
    key = key_file.read()
    cert = cert_file.read()

    if len(key) != 36:
        raise CliException(f"Invalid key length {len(key)} (expected 36)")
    ecdsa_key = SigningKey.from_string(key[4:], curve=NIST256p)
    pem_pubkey = serialization.load_pem_public_key(
        ecdsa_key.get_verifying_key().to_pem()
    )

    x509_cert = x509.load_der_x509_certificate(cert)
    cert_pubkey = x509_cert.public_key()

    if not isinstance(pem_pubkey, EllipticCurvePublicKey):
        raise CliException("The FIDO2 attestation key is not an EC key")
    if not isinstance(cert_pubkey, EllipticCurvePublicKey):
        raise CliException(
            "The FIDO2 attestation certificate does not contain an EC key"
        )
    if pem_pubkey.public_numbers() != cert_pubkey.public_numbers():
        raise CliException(
            "The FIDO2 attestation certificate does not match the public key"
        )

    # See https://www.w3.org/TR/webauthn-2/#sctn-packed-attestation-cert-requirements
    if x509_cert.version != x509.Version.v3:
        raise CliException(
            f"Unexpected certificate version {x509_cert.version} (expected v3)"
        )

    subject_attrs = {
        attr.rfc4514_attribute_name: attr.value for attr in x509_cert.subject
    }
    for name in ["C", "CN", "O", "OU"]:
        if name not in subject_attrs:
            raise CliException(f"Missing subject {name} in certificate")
    if subject_attrs["OU"] != "Authenticator Attestation":
        raise CliException(
            f"Unexpected certificate subject OU {subject_attrs['OU']!r} (expected "
            "Authenticator Attestation)"
        )

    found_aaguid = False
    for extension in x509_cert.extensions:
        if extension.oid.dotted_string == "1.3.6.1.4.1.45724.1.1.4":
            found_aaguid = True
    if not found_aaguid:
        raise CliException("Missing AAGUID extension in certificate")

    basic_constraints = x509_cert.extensions.get_extension_for_class(
        x509.BasicConstraints
    )
    if basic_constraints.value.ca:
        raise CliException("CA must be set to false in the basic constraints")

    cert_hash = sha256(cert).digest().hex()
    print(f"FIDO2 certificate hash: {cert_hash}")

    with ctx.connect_device() as device:
        provisioner = ProvisionerApp(device)
        provisioner.write_file(b"fido/x5c/00", cert)
        provisioner.write_file(b"fido/sec/00", key)


# This import has to be added here to avoid circular dependency
# Import "secrets" subcommand from the secrets module
from . import secrets  # noqa: F401,E402



#
# Copyright (c) 2016 Nordic Semiconductor ASA
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright notice, this
#   list of conditions and the following disclaimer in the documentation and/or
#   other materials provided with the distribution.
#
#   3. Neither the name of Nordic Semiconductor ASA nor the names of other
#   contributors to this software may be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#   4. This software must only be used in or with a processor manufactured by Nordic
#   Semiconductor ASA, or in or with a processor manufactured by a third party that
#   is used in combination with a processor manufactured by Nordic Semiconductor.
#
#   5. Any software provided in binary or object form under this license must not be
#   reverse engineered, decompiled, modified and/or disassembled.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import binascii
import hashlib

# Python standard library
import os
import shutil
import tempfile
from enum import Enum

# 3rd party libraries
from zipfile import ZipFile

from .crc16 import calc_crc16
from .init_packet_pb import (
    CommandTypes,
    DFUType,
    HashTypes,
    InitPacketPB,
    SigningTypes,
    ValidationTypes,
)
from .manifest import Manifest, ManifestGenerator
from .model import FirmwareKeys, HexType

# Nordic libraries
from .nrfhex import nRFHex
from .signing import Signing

# from csemi.zigbee.ota_file import OTA_file


HexTypeToInitPacketFwTypemap = {
    HexType.APPLICATION: DFUType.APPLICATION,
    HexType.BOOTLOADER: DFUType.BOOTLOADER,
    HexType.SOFTDEVICE: DFUType.SOFTDEVICE,
    HexType.SD_BL: DFUType.SOFTDEVICE_BOOTLOADER,
    HexType.EXTERNAL_APPLICATION: DFUType.EXTERNAL_APPLICATION,
}


class PackageException(Exception):
    pass


class PacketField(Enum):
    DEBUG_MODE = 1
    HW_VERSION = 2
    FW_VERSION = 3
    REQUIRED_SOFTDEVICES_ARRAY = 4


class Package:
    """
    Packages and unpacks Nordic DFU packages. Nordic DFU packages are zip files that contains firmware and meta-information
    necessary for utilities to perform a DFU on nRF5X devices.

    The internal data model used in Package is a dictionary. The dictionary is expressed like this in
     json format:

     {
        "manifest": {
            "bootloader": {
                "bin_file": "asdf.bin",
                "dat_file": "asdf.dat",
                "init_packet_data": {
                    "application_version": null,
                    "device_revision": null,
                    "device_type": 5,
                    "firmware_hash": "asdfasdkfjhasdkfjashfkjasfhaskjfhkjsdfhasjkhf",
                    "softdevice_req": [
                        17,
                        18
                    ]
                }
            }
    }

    Attributes application, bootloader, softdevice, softdevice_bootloader shall not be put into the manifest if they are null

    """

    DEFAULT_DEBUG_MODE = False
    DEFAULT_HW_VERSION = 0xFFFFFFFF
    DEFAULT_APP_VERSION = 0xFFFFFFFF
    DEFAULT_BL_VERSION = 0xFFFFFFFF
    DEFAULT_SD_REQ = [0xFFFE]
    DEFAULT_SD_ID = [0xFFFE]
    DEFAULT_DFU_VER = 0.5
    MANIFEST_FILENAME = "manifest.json"
    DEFAULT_BOOT_VALIDATION_TYPE = ValidationTypes.VALIDATE_GENERATED_CRC.name

    def __init__(
        self,
        debug_mode=DEFAULT_DEBUG_MODE,
        hw_version=DEFAULT_HW_VERSION,
        app_version=DEFAULT_APP_VERSION,
        bl_version=DEFAULT_BL_VERSION,
        sd_req=DEFAULT_SD_REQ,
        sd_id=DEFAULT_SD_ID,
        app_fw=None,
        bootloader_fw=None,
        softdevice_fw=None,
        sd_boot_validation=DEFAULT_BOOT_VALIDATION_TYPE,
        app_boot_validation=DEFAULT_BOOT_VALIDATION_TYPE,
        signer=None,
        is_external=False,
        zigbee_format=False,
        manufacturer_id=0,
        image_type=0,
        comment="",
        zigbee_ota_min_hw_version=None,
        zigbee_ota_max_hw_version=None,
    ):

        """
        Constructor that requires values used for generating a Nordic DFU package.

        :param int debug_mode: Debug init-packet field
        :param int hw_version: Hardware version init-packet field
        :param int app_version: App version init-packet field
        :param int bl_version: Bootloader version init-packet field
        :param list sd_req: Softdevice Requirement init-packet field
        :param list sd_id: Softdevice Requirement init-packet field for the Application if softdevice_fw is set
        :param str app_fw: Path to application firmware file
        :param str bootloader_fw: Path to bootloader firmware file
        :param str softdevice_fw: Path to softdevice firmware file
        :param Signing signer: Instance of Signing() for Signing key file (PEM)
        :param int zigbee_ota_min_hw_version: Minimal zigbee ota hardware version
        :param int zigbee_ota_max_hw_version: Maximum zigbee ota hardware version
        :return: None
        """

        init_packet_vars = {}
        if debug_mode is not None:
            init_packet_vars[PacketField.DEBUG_MODE] = debug_mode

        if hw_version is not None:
            init_packet_vars[PacketField.HW_VERSION] = hw_version

        if sd_id is not None:
            init_packet_vars[PacketField.REQUIRED_SOFTDEVICES_ARRAY] = sd_id

        if sd_boot_validation is not None:
            sd_boot_validation_type = [ValidationTypes[sd_boot_validation]]
        else:
            sd_boot_validation_type = [ValidationTypes.VALIDATE_GENERATED_CRC]

        if app_boot_validation is not None:
            app_boot_validation_type = [ValidationTypes[app_boot_validation]]
        else:
            app_boot_validation_type = [ValidationTypes.VALIDATE_GENERATED_CRC]

        self.firmwares_data = {}

        if app_fw:
            firmware_type = (
                HexType.EXTERNAL_APPLICATION if is_external else HexType.APPLICATION
            )
            self.__add_firmware_info(
                firmware_type=firmware_type,
                firmware_version=app_version,
                filename=app_fw,
                boot_validation_type=app_boot_validation_type,
                init_packet_data=init_packet_vars,
            )

        # WARNING
        # Do not move the setting of the `REQUIRED_SOFTDEVICES_ARRAY`
        # field to be `sd_req` to before the `self__.add_firmware_info` call
        # for HexType.EXTERNAL_APPLICATION.
        #
        # When doing a dfu update, the sd_req, specifies that whatever is being
        # updated requires some version of the softdevice. In the case when
        # somebody does a an update with both an application and a softdevice,
        # both sd_req and sd_id are used at the same time. Moving assignment up
        # will cause the versions accepted for the softdevice to also be
        # accepted for the application, which can lead to invalid updates. If
        # the value 0x00 is provided, it can also lead to the softdevice being
        # deleted.
        #
        # Moving was tried https://github.com/NordicSemiconductor/pc-nrfutil/pull/349, but a
        # stable solution is currently favored over one solving this particular
        # issue. Any changes will have to be sufficiently tested to enusre a
        # similar bug has not been introduced.
        if sd_req is not None:
            init_packet_vars[PacketField.REQUIRED_SOFTDEVICES_ARRAY] = sd_req

        if bootloader_fw:
            self.__add_firmware_info(
                firmware_type=HexType.BOOTLOADER,
                firmware_version=bl_version,
                filename=bootloader_fw,
                boot_validation_type=[ValidationTypes.VALIDATE_GENERATED_CRC],
                init_packet_data=init_packet_vars,
            )

        if softdevice_fw:
            self.__add_firmware_info(
                firmware_type=HexType.SOFTDEVICE,
                firmware_version=0xFFFFFFFF,
                filename=softdevice_fw,
                boot_validation_type=sd_boot_validation_type,
                init_packet_data=init_packet_vars,
            )

        assert not signer or isinstance(signer, Signing)
        self.signer = signer

        self.work_dir = None
        self.manifest = None

        if zigbee_format:
            self.is_zigbee = True
            self.image_type = image_type
            self.manufacturer_id = manufacturer_id
            self.comment = comment
            self.zigbee_ota_min_hw_version = zigbee_ota_min_hw_version
            self.zigbee_ota_max_hw_version = zigbee_ota_max_hw_version
        else:
            self.is_zigbee = False
            self.image_type = None
            self.manufacturer_id = None
            self.comment = None

    def __del__(self):
        """
        Destructor removes the temporary working directory
        :return:
        """
        if self.work_dir is not None:
            shutil.rmtree(self.work_dir)
        self.work_dir = None

    def rm_work_dir(self, preserve):
        # Delete the temporary directory
        if self.work_dir is not None:
            if not preserve:
                shutil.rmtree(self.work_dir)

        self.work_dir = None

    def parse_package(self, filename, preserve_work_dir=False):
        self.work_dir = self.__create_temp_workspace()

        self.zip_file = filename
        self.zip_dir = os.path.join(self.work_dir, "unpacked_zip")
        self.manifest = Package.unpack_package(filename, self.zip_dir)

        self.rm_work_dir(preserve_work_dir)

    def image_str(self, index, hex_type, img):
        type_strs = {
            HexType.SD_BL: "sd_bl",
            HexType.SOFTDEVICE: "softdevice",
            HexType.BOOTLOADER: "bootloader",
            HexType.APPLICATION: "application",
            HexType.EXTERNAL_APPLICATION: "external application",
        }

        # parse init packet
        with open(os.path.join(self.zip_dir, img.dat_file), "rb") as imgf:
            initp_bytes = imgf.read()

        initp = InitPacketPB(from_bytes=initp_bytes)

        sd_req = ""
        for x in initp.init_command.sd_req:
            sd_req = sd_req + "0x{0:02X}, ".format(x)

        if len(sd_req) != 0:
            sd_req = sd_req[:-2]

        if initp.packet.HasField("signed_command"):
            cmd = initp.packet.signed_command.command
            signature_type = SigningTypes(
                initp.packet.signed_command.signature_type
            ).name
            signature_hex = binascii.hexlify(initp.packet.signed_command.signature)
        else:
            cmd = initp.packet.command
            signature_type = "UNSIGNED"
            signature_hex = "N/A"

        boot_validation_type = []
        boot_validation_bytes = []
        for x in cmd.init.boot_validation:
            boot_validation_type.append(ValidationTypes(x.type).name)
            boot_validation_bytes.append(binascii.hexlify(x.bytes))

        s = """|
|- Image #{0}:
   |- Type: {1}
   |- Image file: {2}
   |- Init packet file: {3}
      |
      |- op_code: {4}
      |- signature_type: {5}
      |- signature (little-endian): {6}
      |
      |- fw_version: 0x{7:08X} ({7})
      |- hw_version 0x{8:08X} ({8})
      |- sd_req: {9}
      |- type: {10}
      |- sd_size: {11}
      |- bl_size: {12}
      |- app_size: {13}
      |
      |- hash_type: {14}
      |- hash (little-endian): {15}
      |
      |- boot_validation_type: {16}
      |- boot_validation_signature (little-endian): {17}
      |
      |- is_debug: {18}

""".format(
            index,
            type_strs[hex_type],
            img.bin_file,
            img.dat_file,
            CommandTypes(cmd.op_code).name,
            signature_type,
            signature_hex,
            cmd.init.fw_version,
            cmd.init.hw_version,
            sd_req,
            DFUType(cmd.init.type).name,
            cmd.init.sd_size,
            cmd.init.bl_size,
            cmd.init.app_size,
            HashTypes(cmd.init.hash.hash_type).name,
            binascii.hexlify(cmd.init.hash.hash),
            boot_validation_type,
            boot_validation_bytes,
            cmd.init.is_debug,
        )

        return s

    def __str__(self):

        imgs = ""
        i = 0
        if self.manifest.softdevice_bootloader:
            imgs = imgs + self.image_str(
                i, HexType.SD_BL, self.manifest.softdevice_bootloader
            )
            i = i + 1

        if self.manifest.softdevice:
            imgs = imgs + self.image_str(
                i, HexType.SOFTDEVICE, self.manifest.softdevice
            )
            i = i + 1

        if self.manifest.bootloader:
            imgs = imgs + self.image_str(
                i, HexType.BOOTLOADER, self.manifest.bootloader
            )
            i = i + 1

        if self.manifest.application:
            imgs = imgs + self.image_str(
                i, HexType.APPLICATION, self.manifest.application
            )
            i = i + 1

        s = """
DFU Package: <{0}>:
|
|- Image count: {1}
""".format(
            self.zip_file, i
        )

        s = s + imgs
        return s

    def generate_package(self, filename, preserve_work_dir=False):
        """
        Generates a Nordic DFU package. The package is a zip file containing firmware(s) and metadata required
        for Nordic DFU applications to perform DFU onn nRF5X devices.

        :param str filename: Filename for generated package.
        :param bool preserve_work_dir: True to preserve the temporary working directory.
        Useful for debugging of a package, and if the user wants to look at the generated package without having to
        unzip it.
        :return: None
        """
        self.zip_file = filename
        self.work_dir = self.__create_temp_workspace()

        sd_bin_created = False
        if Package._is_bootloader_softdevice_combination(self.firmwares_data):
            # Removing softdevice and bootloader data from dictionary and adding the combined later
            softdevice_fw_data = self.firmwares_data.pop(HexType.SOFTDEVICE)
            bootloader_fw_data = self.firmwares_data.pop(HexType.BOOTLOADER)

            softdevice_fw_name = softdevice_fw_data[FirmwareKeys.FIRMWARE_FILENAME]
            bootloader_fw_name = bootloader_fw_data[FirmwareKeys.FIRMWARE_FILENAME]

            new_filename = "sd_bl.bin"
            sd_bl_file_path = os.path.join(self.work_dir, new_filename)

            nrf_hex = nRFHex(softdevice_fw_name, bootloader_fw_name)
            nrf_hex.tobinfile(sd_bl_file_path)

            softdevice_size = nrf_hex.size()
            bootloader_size = nrf_hex.bootloadersize()

            boot_validation_type = []
            boot_validation_type.extend(
                softdevice_fw_data[FirmwareKeys.BOOT_VALIDATION_TYPE]
            )
            boot_validation_type.extend(
                bootloader_fw_data[FirmwareKeys.BOOT_VALIDATION_TYPE]
            )

            self.__add_firmware_info(
                firmware_type=HexType.SD_BL,
                firmware_version=bootloader_fw_data[FirmwareKeys.INIT_PACKET_DATA][
                    PacketField.FW_VERSION
                ],  # use bootloader version in combination with SD
                filename=sd_bl_file_path,
                init_packet_data=softdevice_fw_data[FirmwareKeys.INIT_PACKET_DATA],
                boot_validation_type=boot_validation_type,
                sd_size=softdevice_size,
                bl_size=bootloader_size,
            )

            # Need to generate SD only bin for boot validation signature
            sd_bin = Package.normalize_firmware_to_bin(
                self.work_dir, softdevice_fw_data[FirmwareKeys.FIRMWARE_FILENAME]
            )
            sd_bin_path = os.path.join(self.work_dir, sd_bin)
            sd_bin_created = True

        for key, firmware_data in self.firmwares_data.items():

            # Normalize the firmware file and store it in the work directory
            firmware_data[
                FirmwareKeys.BIN_FILENAME
            ] = Package.normalize_firmware_to_bin(
                self.work_dir, firmware_data[FirmwareKeys.FIRMWARE_FILENAME]
            )

            # Calculate the hash for the .bin file located in the work directory
            bin_file_path = os.path.join(
                self.work_dir, firmware_data[FirmwareKeys.BIN_FILENAME]
            )
            firmware_hash = Package.calculate_sha256_hash(bin_file_path)
            bin_length = int(Package.calculate_file_size(bin_file_path))

            sd_size = 0
            bl_size = 0
            app_size = 0
            if key in [HexType.APPLICATION, HexType.EXTERNAL_APPLICATION]:
                app_size = bin_length
            elif key == HexType.SOFTDEVICE:
                sd_size = bin_length
            elif key == HexType.BOOTLOADER:
                bl_size = bin_length
            elif key == HexType.SD_BL:
                bl_size = firmware_data[FirmwareKeys.BL_SIZE]
                sd_size = firmware_data[FirmwareKeys.SD_SIZE]

            boot_validation_type_array = firmware_data[
                FirmwareKeys.BOOT_VALIDATION_TYPE
            ]
            boot_validation_bytes_array = []
            for x in boot_validation_type_array:
                if x == ValidationTypes.VALIDATE_ECDSA_P256_SHA256:
                    if key == HexType.SD_BL:
                        boot_validation_bytes_array.append(
                            Package.sign_firmware(self.signer, sd_bin_path)
                        )
                    else:
                        boot_validation_bytes_array.append(
                            Package.sign_firmware(self.signer, bin_file_path)
                        )
                else:
                    boot_validation_bytes_array.append(b"")

            init_packet = InitPacketPB(
                from_bytes=None,
                hash_bytes=firmware_hash,
                hash_type=HashTypes.SHA256,
                boot_validation_type=boot_validation_type_array,
                boot_validation_bytes=boot_validation_bytes_array,
                dfu_type=HexTypeToInitPacketFwTypemap[key],
                is_debug=firmware_data[FirmwareKeys.INIT_PACKET_DATA][
                    PacketField.DEBUG_MODE
                ],
                fw_version=firmware_data[FirmwareKeys.INIT_PACKET_DATA][
                    PacketField.FW_VERSION
                ],
                hw_version=firmware_data[FirmwareKeys.INIT_PACKET_DATA][
                    PacketField.HW_VERSION
                ],
                sd_size=sd_size,
                app_size=app_size,
                bl_size=bl_size,
                sd_req=firmware_data[FirmwareKeys.INIT_PACKET_DATA][
                    PacketField.REQUIRED_SOFTDEVICES_ARRAY
                ],
            )

            if self.signer is not None:
                signature = self.signer.sign(init_packet.get_init_command_bytes())
                init_packet.set_signature(signature, SigningTypes.ECDSA_P256_SHA256)

            # Store the .dat file in the work directory
            init_packet_filename = firmware_data[FirmwareKeys.BIN_FILENAME].replace(
                ".bin", ".dat"
            )

            with open(
                os.path.join(self.work_dir, init_packet_filename), "wb"
            ) as init_packet_file:
                init_packet_file.write(init_packet.get_init_packet_pb_bytes())

            firmware_data[FirmwareKeys.DAT_FILENAME] = init_packet_filename

            if self.is_zigbee:
                firmware_version = firmware_data[FirmwareKeys.INIT_PACKET_DATA][
                    PacketField.FW_VERSION
                ]
                file_name = firmware_data[FirmwareKeys.BIN_FILENAME]

                self.zigbee_ota_file = OTA_file(
                    firmware_version,
                    len(init_packet.get_init_packet_pb_bytes()),
                    binascii.crc32(init_packet.get_init_packet_pb_bytes()) & 0xFFFFFFFF,
                    init_packet.get_init_packet_pb_bytes(),
                    os.path.getsize(file_name),
                    self.calculate_crc(32, file_name) & 0xFFFFFFFF,
                    bytes(open(file_name, "rb").read()),
                    self.manufacturer_id,
                    self.image_type,
                    self.comment,
                    self.zigbee_ota_min_hw_version,
                    self.zigbee_ota_max_hw_version,
                )

                ota_file_handle = open(self.zigbee_ota_file.filename, "wb")
                ota_file_handle.write(self.zigbee_ota_file.binary)
                ota_file_handle.close()

        # Remove SD binary file created for boot validation
        if sd_bin_created:
            os.remove(sd_bin_path)

        # Store the manifest to manifest.json
        manifest = self.create_manifest()

        with open(
            os.path.join(self.work_dir, Package.MANIFEST_FILENAME), "w"
        ) as manifest_file:
            manifest_file.write(manifest)

        # Package the work_dir to a zip file
        Package.create_zip_package(self.work_dir, filename)

        # Delete the temporary directory
        self.rm_work_dir(preserve_work_dir)

    @staticmethod
    def __create_temp_workspace():
        return tempfile.mkdtemp(prefix="nrf_dfu_pkg_")

    @staticmethod
    def create_zip_package(work_dir, filename):
        files = os.listdir(work_dir)

        with ZipFile(filename, "w") as package:
            for _file in files:
                file_path = os.path.join(work_dir, _file)
                package.write(file_path, _file)

    @staticmethod
    def calculate_file_size(firmware_filename):
        b = os.path.getsize(firmware_filename)
        return b

    @staticmethod
    def calculate_sha256_hash(firmware_filename):
        read_buffer = 4096

        digest = hashlib.sha256()

        with open(firmware_filename, "rb") as firmware_file:
            while True:
                data = firmware_file.read(read_buffer)

                if data:
                    digest.update(data)
                else:
                    break

        # return hash in little endian
        sha256 = digest.digest()
        return sha256[31::-1]

    @staticmethod
    def calculate_crc(crc, firmware_filename):
        """
        Calculates CRC16 has on provided firmware filename

        :type str firmware_filename:
        """
        data_buffer = b""
        read_size = 4096

        with open(firmware_filename, "rb") as firmware_file:
            while True:
                data = firmware_file.read(read_size)

                if data:
                    data_buffer += data
                else:
                    break
        if crc == 16:
            return calc_crc16(data_buffer, 0xFFFF)
        elif crc == 32:
            return binascii.crc32(data_buffer)
        else:
            raise ValueError("Invalid CRC type")

    @staticmethod
    def sign_firmware(signer, firmware_filename):
        assert isinstance(signer, Signing)
        data_buffer = b""
        with open(firmware_filename, "rb") as firmware_file:
            data_buffer = firmware_file.read()
        return signer.sign(data_buffer)

    def create_manifest(self):
        manifest = ManifestGenerator(self.firmwares_data)
        return manifest.generate_manifest()

    @staticmethod
    def _is_bootloader_softdevice_combination(firmwares):
        return (HexType.BOOTLOADER in firmwares) and (HexType.SOFTDEVICE in firmwares)

    def __add_firmware_info(
        self,
        firmware_type,
        firmware_version,
        filename,
        init_packet_data,
        boot_validation_type,
        sd_size=None,
        bl_size=None,
    ):
        self.firmwares_data[firmware_type] = {
            FirmwareKeys.FIRMWARE_FILENAME: filename,
            FirmwareKeys.INIT_PACKET_DATA: init_packet_data.copy(),
            # Copying init packet to avoid using the same for all firmware
            FirmwareKeys.BOOT_VALIDATION_TYPE: boot_validation_type,
        }

        if firmware_type == HexType.SD_BL:
            self.firmwares_data[firmware_type][FirmwareKeys.SD_SIZE] = sd_size
            self.firmwares_data[firmware_type][FirmwareKeys.BL_SIZE] = bl_size

        if firmware_version is not None:
            self.firmwares_data[firmware_type][FirmwareKeys.INIT_PACKET_DATA][
                PacketField.FW_VERSION
            ] = firmware_version

    @staticmethod
    def normalize_firmware_to_bin(work_dir, firmware_path):
        firmware_filename = os.path.basename(firmware_path)
        new_filename = firmware_filename.replace(".hex", ".bin")
        new_filepath = os.path.join(work_dir, new_filename)

        if not os.path.exists(new_filepath):
            temp = nRFHex(firmware_path)
            temp.tobinfile(new_filepath)

        return new_filepath

    @staticmethod
    def unpack_package(package_path, target_dir):
        """
        Unpacks a Nordic DFU package.

        :param str package_path: Path to the package
        :param str target_dir: Target directory to unpack the package to
        :return: Manifest Manifest: Returns a manifest back to the user. The manifest is a parse datamodel
        of the manifest found in the Nordic DFU package.
        """

        if not os.path.isfile(package_path):
            raise PackageException("Package {0} not found.".format(package_path))

        target_dir = os.path.abspath(target_dir)
        target_base_path = os.path.dirname(target_dir)

        if not os.path.exists(target_base_path):
            raise PackageException(
                "Base path to target directory {0} does not exist.".format(
                    target_base_path
                )
            )

        if not os.path.isdir(target_base_path):
            raise PackageException(
                "Base path to target directory {0} is not a directory.".format(
                    target_base_path
                )
            )

        if os.path.exists(target_dir):
            raise PackageException(
                "Target directory {0} exists, not able to unpack to that directory.",
                target_dir,
            )

        with ZipFile(package_path, "r") as pkg:
            pkg.extractall(target_dir)

            with open(os.path.join(target_dir, Package.MANIFEST_FILENAME), "r") as f:
                _json = f.read()
                """:type :str """

                return Manifest.from_json(_json)



#
# Copyright (c) 2019 Nordic Semiconductor ASA
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright notice, this
#   list of conditions and the following disclaimer in the documentation and/or
#   other materials provided with the distribution.
#
#   3. Neither the name of Nordic Semiconductor ASA nor the names of other
#   contributors to this software may be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#   4. This software must only be used in or with a processor manufactured by Nordic
#   Semiconductor ASA, or in or with a processor manufactured by a third party that
#   is used in combination with a processor manufactured by Nordic Semiconductor.
#
#   5. Any software provided in binary or object form under this license must not be
#   reverse engineered, decompiled, modified and/or disassembled.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#


import ctypes
import logging
import os
import sys
from importlib import import_module

# from pc_ble_driver_py.exceptions import NordicSemiException
from ..exceptions import NordicSemiException

LIBUSB_ENDPOINT_IN = 0x80
LIBUSB_ENDPOINT_OUT = 0x00
LIBUSB_REQUEST_TYPE_STANDARD = 0x00 << 5
LIBUSB_REQUEST_TYPE_CLASS = 0x01 << 5
LIBUSB_REQUEST_TYPE_VENDOR = 0x02 << 5
LIBUSB_REQUEST_TYPE_RESERVED = 0x03 << 5
LIBUSB_RECIPIENT_DEVICE = 0x00
LIBUSB_RECIPIENT_INTERFACE = 0x01
LIBUSB_RECIPIENT_ENDPOINT = 0x02
LIBUSB_RECIPIENT_OTHER = 0x03

ReqTypeInterfaceClass = LIBUSB_REQUEST_TYPE_CLASS | LIBUSB_RECIPIENT_INTERFACE
ReqTypeIN = ReqTypeInterfaceClass | LIBUSB_ENDPOINT_IN
ReqTypeOUT = ReqTypeInterfaceClass | LIBUSB_ENDPOINT_OUT
NORDIC_SEM_VER_REQUEST = 8
NORDIC_DFU_INFO_REQUEST = 7
DFU_DETACH_REQUEST = 0

logger = logging.getLogger(__name__)

is_32_bit = ctypes.sizeof(ctypes.c_void_p) == 4
abs_file_dir = os.path.dirname(os.path.abspath(__file__))
rel_import_dir = ""

dfu_path = os.path.join("nordicsemi", "dfu")
if is_32_bit:
    libusb_path = os.path.join("libusb", "x86")
    abs_file_dir = abs_file_dir.replace(dfu_path, libusb_path)
    rel_import_dir = os.path.join(".", libusb_path)
else:
    libusb_path = os.path.join("libusb", "x64")
    abs_file_dir = abs_file_dir.replace(dfu_path, libusb_path)
    rel_import_dir = os.path.join(".", libusb_path)

for path in [
    "PATH",
    "LD_LIBRARY_PATH",
    "DYLD_LIBRARY_PATH",
    "DYLD_FALLBACK_LIBRARY_PATH",
]:
    if path not in os.environ:
        os.environ[path] = ""
    os.environ[path] = (
        rel_import_dir + os.pathsep + abs_file_dir + os.pathsep + os.environ[path]
    )


class DFUTrigger:
    def __init__(self):
        self.context = None

        if sys.platform == "win32":
            try:
                # Load the libusb dll in advance so that libusb1 module is better able to load it
                ctypes.CDLL(os.path.join(abs_file_dir, "libusb-1.0.dll"))
            except OSError as err:
                logger.info(err)

        try:
            self.usb1 = import_module("usb1")
            self.context = self.usb1.USBContext()
        except OSError as err:
            if "libusb" in str(err):
                if sys.platform == "win32" or sys.platform == "darwin":
                    show_msg = (
                        "Libusb1 binaries are bundled with nrfutil for Windows and MacOS. "
                        "Python is unable to locate or load the binaries. "
                    )
                elif "linux" in sys.platform:
                    show_msg = (
                        "Libusb1 binaries are bundled with some linux distributions. "
                        "If you see this message, they are probably not installed on your system. "
                        "If you want to use DFU trigger, please install 'libusb1' using your package manager. "
                        "E.g: 'sudo apt-get install libusb-dev'."
                    )
                else:
                    show_msg = "Libusb1 is not compatible with your operating system."

                logger.warning(
                    "Could not load libusb1-0 library, which is a requirement to use DFU trigger. "
                    "This is not a problem unless you intend to use this functionality. "
                    "{}".format(show_msg)
                )

    def clean(self):
        self.usb1 = None
        if self.context:
            self.context.close()

    def select_device(self, listed_device):
        all_devices = self.context.getDeviceList()
        filtered_devices = [
            dev
            for dev in all_devices
            if hex(dev.getVendorID())[2:].lower() == listed_device.vendor_id.lower()
            and hex(dev.getProductID())[2:].lower() == listed_device.product_id.lower()
        ]

        access_error = False
        triggerless_devices = 0

        for nordic_device in filtered_devices:
            try:
                handle = nordic_device.open()
                SNO = handle.getSerialNumber()
                handle.close()
                if SNO.lower() == listed_device.serial_number.lower():
                    return nordic_device

            except self.usb1.USBErrorNotFound as err:
                #  Devices with matching VID and PID as target, but without a trigger iface.
                triggerless_devices += 1
            except self.usb1.USBErrorAccess as err:
                access_error = True
            except self.usb1.USBErrorNotSupported as err:
                pass  #  Unsupported device. Moving on

        if triggerless_devices > 0:
            logger.debug(
                "DFU trigger: Could not find trigger interface for device with serial number {}. "
                "{}/{} devices with same VID/PID were missing a trigger interface.".format(
                    listed_device.serial_number,
                    triggerless_devices,
                    len(filtered_devices),
                )
            )

        if access_error:
            raise NordicSemiException(
                "LIBUSB_ERROR_ACCESS: Unable to connect to trigger interface."
            )

    def get_dfu_interface_num(self, libusb_device):
        for setting in libusb_device.iterSettings():
            if (
                setting.getClass() == 255
                and setting.getSubClass() == 1
                and setting.getProtocol() == 1
            ):
                return setting.getNumber()

    def no_trigger_exception(self, device):
        return NordicSemiException(
            "No trigger interface found for device with serial number: {}, Product ID: 0x{} and Vendor ID: 0x{}\n".format(
                device.serial_number, device.product_id, device.vendor_id
            )
        )

    def enter_bootloader_mode(self, listed_device):
        if self.context is None:
            raise NordicSemiException(
                "No Libusb1 context found, but is required to use DFU trigger. "
                "This likely happens because the libusb1-0 binaries are missing from your system, "
                "or Python is unable to locate them."
            )
        libusb_device = self.select_device(listed_device)
        if libusb_device is None:
            raise self.no_trigger_exception(listed_device)
        device_handle = libusb_device.open()
        dfu_iface = self.get_dfu_interface_num(libusb_device)

        if dfu_iface is None:
            raise self.no_trigger_exception(listed_device)

        with device_handle.claimInterface(dfu_iface):
            arr = bytearray("0", "utf-8")
            try:
                device_handle.controlWrite(
                    ReqTypeOUT, DFU_DETACH_REQUEST, 0, dfu_iface, arr
                )
            except Exception as err:
                if "LIBUSB_ERROR_PIPE" in str(err):
                    return
        raise NordicSemiException(
            "A diconnection event from libusb is expected when the usb device restarts after triggering bootloder. "
            "The event was not received. This can be an indication that the device was unable to leave application mode. "
            "Serial number: {}, Product ID: 0x{}, Vendor ID: 0x{}\n\n".format(
                listed_device.serial_number,
                listed_device.product_id,
                listed_device.vendor_id,
            )
        )



# -*- coding: utf-8 -*-
#
# Copyright 2021-2022 Nitrokey Developers
#
# Licensed under the Apache License, Version 2.0, <LICENSE-APACHE or
# http://apache.org/licenses/LICENSE-2.0> or the MIT license <LICENSE-MIT or
# http://opensource.org/licenses/MIT>, at your option. This file may not be
# copied, modified, or distributed except according to those terms.

import logging
from contextlib import contextmanager
from typing import Any, Callable, Iterator, List, Optional

from click import Abort

from pynitrokey.cli.exceptions import CliException
from pynitrokey.cli.nk3 import Context
from pynitrokey.helpers import DownloadProgressBar, ProgressBar, confirm, local_print
from pynitrokey.nk3.updates import Updater, UpdateUi
from pynitrokey.nk3.utils import Version

logger = logging.getLogger(__name__)


class UpdateCli(UpdateUi):
    def __init__(self) -> None:
        self._version_printed = False

    def error(self, *msgs: Any) -> Exception:
        return CliException(*msgs)

    def abort(self, *msgs: Any) -> Exception:
        return CliException(*msgs, support_hint=False)

    def abort_downgrade(self, current: Version, image: Version) -> Exception:
        self._print_firmware_versions(current, image)
        return self.abort(
            "The firmware image is older than the firmware on the device."
        )

    def abort_pynitrokey_version(
        self, current: Version, required: Version
    ) -> Exception:
        return self.abort(
            f"This update requires pynitrokey version {required} (current: {current}). "
            "Please update pynitrokey to install the update."
        )

    def confirm_download(self, current: Optional[Version], new: Version) -> None:
        confirm(
            f"Do you want to download the firmware version {new}?",
            default=True,
            abort=True,
        )

    def confirm_pynitrokey_version(self, current: Version, required: Version) -> None:
        local_print(
            f"This update requires pynitrokey version {required} (current: {current})."
        )
        local_print("Using an outdated pynitrokey version is strongly discouraged.")
        if not confirm(
            "Do you want to continue with an outdated pynitrokey version at your own risk?"
        ):
            logger.info("Update cancelled by user")
            raise Abort()

    def confirm_update(self, current: Optional[Version], new: Version) -> None:
        self._print_firmware_versions(current, new)
        local_print("")
        local_print(
            "Please do not remove the Nitrokey 3 or insert any other Nitrokey 3 devices "
            "during the update. Doing so may damage the Nitrokey 3."
        )
        if not confirm("Do you want to perform the firmware update now?"):
            logger.info("Update cancelled by user")
            raise Abort()

    def confirm_update_same_version(self, version: Version) -> None:
        self._print_firmware_versions(version, version)
        if not confirm(
            "The version of the firmware image is the same as on the device.  Do you want "
            "to continue anyway?"
        ):
            raise Abort()

    def confirm_extra_information(self, txt: List[str]) -> None:
        if txt:
            local_print("\n".join(txt))
            if not confirm("Have you read these information? Do you want to continue?"):
                raise Abort()

    def request_repeated_update(self) -> Exception:
        local_print(
            "Bootloader mode enabled. Please repeat this command to apply the update."
        )
        return Abort()

    def request_bootloader_confirmation(self) -> None:
        local_print("")
        local_print(
            "Please press the touch button to reboot the device into bootloader mode ..."
        )
        local_print("")

    @contextmanager
    def download_progress_bar(self, desc: str) -> Iterator[Callable[[int, int], None]]:
        with DownloadProgressBar(desc) as bar:
            yield bar.update

    @contextmanager
    def update_progress_bar(self) -> Iterator[Callable[[int, int], None]]:
        with ProgressBar(
            desc="Perform firmware update", unit="B", unit_scale=True
        ) as bar:
            yield bar.update_sum

    @contextmanager
    def finalization_progress_bar(self) -> Iterator[Callable[[int, int], None]]:
        with ProgressBar(desc="Finalize upgrade", unit="%", unit_scale=False) as bar:
            yield bar.update_sum

    def _print_firmware_versions(
        self, current: Optional[Version], new: Optional[Version]
    ) -> None:
        if not self._version_printed:
            current_str = str(current) if current else "[unknown]"
            local_print(f"Current firmware version:  {current_str}")
            local_print(f"Updated firmware version:  {new}")
            self._version_printed = True


def update(
    ctx: Context,
    image: Optional[str],
    version: Optional[str],
    ignore_pynitrokey_version: bool,
) -> Version:
    with ctx.connect() as device:
        updater = Updater(UpdateCli(), ctx.await_bootloader, ctx.await_device)
        return updater.update(device, image, version, ignore_pynitrokey_version)



# -*- coding: utf-8 -*-
#
# Copyright 2021 Nitrokey Developers
#
# Licensed under the Apache License, Version 2.0, <LICENSE-APACHE or
# http://apache.org/licenses/LICENSE-2.0> or the MIT license <LICENSE-MIT or
# http://opensource.org/licenses/MIT>, at your option. This file may not be
# copied, modified, or distributed except according to those terms.

import enum
import logging
import platform
import sys
from enum import Enum
from typing import List, Optional

from fido2.ctap import CtapError
from fido2.hid import CtapHidDevice, open_device

from pynitrokey.fido2 import device_path_to_str

from .base import Nitrokey3Base
from .exceptions import TimeoutException
from .utils import Uuid, Version

RNG_LEN = 57
UUID_LEN = 16
VERSION_LEN = 4

logger = logging.getLogger(__name__)


@enum.unique
class Command(Enum):
    """Vendor-specific CTAPHID commands for the Nitrokey 3."""

    UPDATE = 0x51
    REBOOT = 0x53
    RNG = 0x60
    VERSION = 0x61
    UUID = 0x62
    LOCKED = 0x63
    OTP = 0x70
    PROVISIONER = 0x71
    ADMIN = 0x72


@enum.unique
class BootMode(Enum):
    FIRMWARE = enum.auto()
    BOOTROM = enum.auto()


class Nitrokey3Device(Nitrokey3Base):
    """A Nitrokey 3 device running the firmware."""

    def __init__(self, device: CtapHidDevice) -> None:
        from . import PID_NITROKEY3_DEVICE, VID_NITROKEY
        from .admin_app import AdminApp

        (vid, pid) = (device.descriptor.vid, device.descriptor.pid)
        if (vid, pid) != (VID_NITROKEY, PID_NITROKEY3_DEVICE):
            raise ValueError(
                "Not a Nitrokey 3 device: expected VID:PID "
                f"{VID_NITROKEY:x}:{PID_NITROKEY3_DEVICE:x}, got {vid:x}:{pid:x}"
            )

        self.device = device
        self._path = device_path_to_str(device.descriptor.path)
        self.logger = logger.getChild(self._path)

        self.admin = AdminApp(self)
        self.admin.status()

    @property
    def path(self) -> str:
        return self._path

    @property
    def name(self) -> str:
        return "Nitrokey 3"

    def close(self) -> None:
        self.device.close()

    def reboot(self, mode: BootMode = BootMode.FIRMWARE) -> bool:
        try:
            if mode == BootMode.FIRMWARE:
                self._call(Command.REBOOT)
            elif mode == BootMode.BOOTROM:
                try:
                    self._call(Command.UPDATE)
                except CtapError as e:
                    # The admin app returns an Invalid Length error if the user confirmation
                    # request times out
                    if e.code == CtapError.ERR.INVALID_LENGTH:
                        raise TimeoutException()
                    else:
                        raise e
        except OSError as e:
            # OS error is expected as the device does not respond during the reboot
            self.logger.debug("ignoring OSError after reboot", exc_info=e)
        return True

    def uuid(self) -> Optional[Uuid]:
        uuid = self._call(Command.UUID)
        if len(uuid) == 0:
            # Firmware version 1.0.0 does not support querying the UUID
            return None
        if len(uuid) != UUID_LEN:
            raise ValueError(f"UUID response has invalid length {len(uuid)}")
        return Uuid(int.from_bytes(uuid, byteorder="big"))

    def version(self) -> Version:
        return self.admin.version()

    def factory_reset(self) -> None:
        self.admin.factory_reset()

    def factory_reset_app(self, app: str) -> None:
        self.admin.factory_reset_app(app)

    def wink(self) -> None:
        self.device.wink()

    def rng(self) -> bytes:
        return self._call(Command.RNG, response_len=RNG_LEN)

    def otp(self, data: bytes = b"") -> bytes:
        return self._call(Command.OTP, data=data)

    def is_locked(self) -> bool:
        response = self._call(Command.LOCKED, response_len=1)
        return response[0] == 1

    def _call(
        self, command: Command, response_len: Optional[int] = None, data: bytes = b""
    ) -> bytes:
        response = self.device.call(command.value, data=data)
        if response_len is not None and response_len != len(response):
            raise ValueError(
                f"The response for the CTAPHID {command.name} command has an unexpected length "
                f"(expected: {response_len}, actual: {len(response)})"
            )
        return response

    @staticmethod
    def list() -> List["Nitrokey3Device"]:
        devices = []
        for device in CtapHidDevice.list_devices():
            try:
                devices.append(Nitrokey3Device(device))
            except ValueError:
                # not a Nitrokey 3 device, skip
                pass
        return devices

    @staticmethod
    def open(path: str) -> Optional["Nitrokey3Device"]:
        try:
            if platform.system() == "Windows":
                device = open_device(bytes(path, "utf-8"))
            else:
                device = open_device(path)
        except Exception:
            logger.warn(f"No CTAPHID device at path {path}", exc_info=sys.exc_info())
            return None
        try:
            return Nitrokey3Device(device)
        except ValueError:
            logger.warn(f"No Nitrokey 3 device at path {path}", exc_info=sys.exc_info())
            return None



import logging
import os
import tempfile
from enum import IntEnum


class Verbosity(IntEnum):
    """regular lvls from `logging` & `machine` for machine-readable output only"""

    machine = 100
    silent = logging.CRITICAL
    minimal = logging.WARNING
    user = logging.INFO
    debug = logging.DEBUG
    unset = logging.NOTSET


ENV_DEBUG_VAR = "PYNK_DEBUG"
DEFAULT_VERBOSE = Verbosity.user
VERBOSE = DEFAULT_VERBOSE

# set global debug/verbosity -> search for environment variable: ENV_DEBUG_VAR
_env_dbg_lvl = os.environ.get(ENV_DEBUG_VAR)
if _env_dbg_lvl:
    try:
        # env-var only set w/o contents equals 'Verbosity.debug'
        if _env_dbg_lvl.strip() == "":
            VERBOSE = Verbosity.debug
        # non-empty env-var shall be a number, representing a level
        else:
            VERBOSE = Verbosity(int(_env_dbg_lvl))
    except ValueError as e:
        VERBOSE = DEFAULT_VERBOSE
        print(f"exception: {e}")
        print(
            f"environment variable: '{ENV_DEBUG_VAR}' invalid, "
            f"setting default: {VERBOSE.name} = {VERBOSE.value}"
        )

LOG_FN = tempfile.NamedTemporaryFile(prefix="nitropy.log.").name
LOG_FORMAT_STDOUT = "%(asctime)-15s %(levelname)6s %(name)10s %(message)s"
LOG_FORMAT = "%(relativeCreated)-8d %(levelname)6s %(name)10s %(message)s"

CLI_LOG_BLACKLIST: dict[str, int] = {
    # dict of {name: lenght} mapping to exclude from cli parameter logging
    # name: name of the parameter
    # length: number of arguments the parameter has
    # nitropy start kdf-details
    "--passwd": 1,
    # nitropy start update
    # nitropy pro enable-update
    # nitropy nethsm
    "-p": 1,
    # nitropy pro enable-update
    # nitropy nk3 secrets set-pin
    # nitropy nethsm
    "--password": 1,
    # nitropy nk3 test
    # nitropy fido2 verify
    # nitropy fido2 make-credential
    # nitropy fido2 list-credentials
    # nitropy fido2 delete-credential
    # nitropy fido2 challenge-response
    "--pin": 1,
}

GH_ISSUES_URL = "https://github.com/Nitrokey/pynitrokey/issues/"
SUPPORT_URL = "https://support.nitrokey.com/"
SUPPORT_EMAIL = "support@nitrokey.com"
UDEV_URL = (
    "https://docs.nitrokey.com/nitrokey3/linux/firmware-update.html#troubleshooting"
)
